################################################################################
#
# Copyright (C) 2016-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import os
import warnings
from argparse import Action, ArgumentParser
from typing import Any, Dict, List, Optional

from ..Common import DeveloperWarning, architectureMap
from ..Utilities.Toolchain import ToolchainDefaults, validateToolchain


class DeprecatedOption(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        warnings.warn(
            f"[DEPRECATED] The option {option_string} will be removed in future versions."
        )


def splitExtraParameters(par):
    """Allows the --global-parameters option to specify any parameters from the command line."""

    (key, value) = par.split("=")
    value = eval(value)
    return (key, value)


def parseArguments(input: Optional[List[str]] = None) -> Dict[str, Any]:
    """Parse command line arguments for TensileCreateLibrary.

    Args:
        input: List of strings representing command line arguments used when
               calling parseArguments prgrammatically e.g. in testing.

    Returns:
        A dictionary containing the keys representing options and their values.
    """

    parser = ArgumentParser(
        description="TensileCreateLibrary generates libraries and code object files "
        "for a set of supplied logic files.",
    )

    # Positional arguments
    parser.add_argument("LogicPath", help="Path to Library Logic (YAML) files.")
    parser.add_argument("OutputPath", help="Build path for library files.")
    parser.add_argument(
        "RuntimeLanguage",
        help="Runtime language for generated library.",
        choices=["HIP", "HSA"],
    )

    # Optional arguments
    parser.add_argument(
        "--cxx-compiler",
        dest="CxxCompiler",
        default=ToolchainDefaults.CXX_COMPILER,
        type=str,
        help="C++ compiler used when generating binaries."
        "On Linux, amdclang++ (default) or hipcc. On Windows clang++ (default) or hipcc. "
        "On Windows, include the file extension, or extensions will be searched according to the PATHEXT environment variable. "
        "Pass a fully-qualified path to override environment inspection when searching for the compiler.",
    )
    parser.add_argument(
        "--c-compiler",
        dest="CCompiler",
        default=ToolchainDefaults.C_COMPILER,
        type=str,
    )
    parser.add_argument(
        "--assembler",
        dest="Assembler",
        default=ToolchainDefaults.ASSEMBLER,
        type=str,
    )
    parser.add_argument(
        "--offload-bundler",
        dest="OffloadBundler",
        default=ToolchainDefaults.OFFLOAD_BUNDLER,
        type=str,
    )
    parser.add_argument(
        "--architecture",
        dest="Architecture",
        default="all",
        type=str,
        help="Architectures to generate a library for. When specifying multiple options, "
        "use quoted, semicolon delimited architectures, e.g., --architecture='gfx908;gfx1012'. "
        "Supported archiectures include: " + ", ".join(architectureMap.keys()),
    )
    parser.add_argument(
        "--code-object-version",
        dest="CodeObjectVersion",
        choices=["default", "V4", "V5"],
        type=str,
        help="HSA code-object version.",
    )
    parser.add_argument(
        "--cmake-cxx-compiler",
        dest="CmakeCxxCompiler",
        action=DeprecatedOption,
        help="(Deprecated) Set the environment variable CMAKE_CXX_COMPILER.",
    )

    # Boolean flags
    parser.add_argument(
        "--merge-files",
        dest="MergeFiles",
        action="store_true",
        help="Store all solutions in single file.",
    )
    parser.add_argument(
        "--no-merge-files",
        dest="MergeFiles",
        action="store_false",
        help="Store every solution and kernel in separate file.",
    )
    parser.add_argument(
        "--num-merged-files",
        dest="NumMergedFiles",
        default=1,
        type=int,
        help="Number of files the kernels should be written into.",
    )
    parser.add_argument(
        "--short-file-names",
        dest="ShortNames",
        action="store_true",
        help="Converts solution and kernel names to serial IDs.",
    )
    parser.add_argument(
        "--no-short-file-names",
        dest="ShortNames",
        action="store_false",
        help="Disables short files names.",
    )
    parser.add_argument(
        "--no-enumerate",
        dest="NoEnumerate",
        action="store_true",
        help="Do not run rocm_agent_enumerator.",
    )
    parser.add_argument(
        "--embed-library",
        dest="EmbedLibrary",
        type=str,
        help="Embed (new) library files into static variables. Specify the name of the library",
    )
    parser.add_argument(
        "--embed-library-key",
        dest="EmbedLibraryKey",
        default=None,
        help="Access key for embedding library files.",
    )
    parser.add_argument(
        "--version",
        dest="Version",
        type=str,
        help="Version string to embed into library file.",
    )
    parser.add_argument(
        "--generate-manifest-and-exit",
        dest="GenerateManifestAndExit",
        action="store_true",
        default=False,
        help="Output manifest file with list of expected library objects and exit.",
    )
    parser.add_argument(
        "--generate-sources-and-exit",
        dest="GenerateSourcesAndExit",
        action="store_true",
        default=False,
        help="Output source files only and exit.",
    )
    parser.add_argument(
        "--verify-manifest",
        dest="VerifyManifest",
        action="store_true",
        default=False,
        help="Verify manifest file against generated library files and exit.",
    )
    parser.add_argument(
        "--keep-build-tmp",
        dest="KeepBuildTmp",
        action="store_true",
        default=False,
        help="Do not remove the temporary build directory (may required hundreds of GBs of space)",
    )
    parser.add_argument(
        "--library-format",
        dest="LibraryFormat",
        choices=["yaml", "msgpack"],
        action="store",
        default="msgpack",
        help="Select which library format to use.",
    )
    parser.add_argument(
        "--jobs",
        "-j",
        dest="CpuThreads",
        default=-1,
        type=int,
        help="Number of parallel jobs to launch. "
        "If jobs < 1 or jobs > nproc the number of parallel jobs will be set to the "
        "number of cores, up to a maximum of 64.",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        dest="PrintLevel",
        choices=[0, 1, 2, 3],
        default=1,
        type=int,
        help="Set printing verbosity level.",
    )
    parser.add_argument(
        "--separate-architectures",
        dest="SeparateArchitectures",
        action="store_true",
        default=False,
        help="Separates generated library files by architecture.",
    )
    parser.add_argument(
        "--lazy-library-loading",
        dest="LazyLibraryLoading",
        action="store_true",
        default=False,
        help="Loads libraries when needed instead of eagerly.",
    )
    parser.add_argument(
        "--client-config",
        dest="ClientConfig",
        action="store_true",
        help="Creates best-solution.ini in the output directory for the library and "
        "code object files.",
    )
    parser.add_argument(
        "--ignore-asm-cap-cache",
        dest="IgnoreAsmCapCache",
        action="store_true",
        default=False,
        help="Ignore ASM capabilities cache and derive the capabilities at runtime.",
    )
    parser.add_argument(
        "--write-master-solution-index",
        dest="WriteMasterSolutionIndex",
        action="store_true",
        default=False,
        help="Output master solution index in csv format including number of kernels per architecture.",
    )
    parser.add_argument(
        "--global-parameters",
        nargs="+",
        dest="GlobalParameters",
        type=splitExtraParameters,
        default=[],
        action=DeprecatedOption,
        help="(Deprecated) Additional global parameters.",
    )
    args = parser.parse_args(input)

    arguments = {
        "LogicPath": args.LogicPath,
        "OutputPath": args.OutputPath,
        "RuntimeLanguage": args.RuntimeLanguage,
        "CxxCompiler": args.CxxCompiler,
        "CCompiler": args.CCompiler,
        "Assembler": args.Assembler,
        "OffloadBundler": args.OffloadBundler,
        "CmakeCxxCompiler": args.CmakeCxxCompiler,
        "CodeObjectVersion": args.CodeObjectVersion,
        "Architecture": args.Architecture,
        "MergeFiles": args.MergeFiles,
        "NumMergedFiles": args.NumMergedFiles,
        "ShortNames": args.ShortNames,
        "CodeFromFiles": False,
        "EmbedLibrary": args.EmbedLibrary,
        "EmbedLibraryKey": args.EmbedLibraryKey,
        "Version": args.Version,
        "LibraryFormat": args.LibraryFormat,
        "GenerateManifestAndExit": args.GenerateManifestAndExit,
        "VerifyManifest": args.VerifyManifest,
        "GenerateSourcesAndExit": args.GenerateSourcesAndExit,
        "CpuThreads": args.CpuThreads,
        "PrintLevel": args.PrintLevel,
        "SeparateArchitectures": args.SeparateArchitectures,
        "LazyLibraryLoading": args.LazyLibraryLoading,
        "ClientConfig": args.ClientConfig,
        "IgnoreAsmCapCache": args.IgnoreAsmCapCache,
        "WriteMasterSolutionIndex": args.WriteMasterSolutionIndex,
        "KeepBuildTmp": args.KeepBuildTmp,
    }

    if args.CmakeCxxCompiler:
        os.environ["CMAKE_CXX_COMPILER"] = args.CmakeCxxCompiler
    if args.GenerateSourcesAndExit:
        # Generated sources are preserved and go into output directory
        arguments["WorkingPath"] = arguments["OutputPath"]
    if args.PrintLevel <= 0:
        warnings.filterwarnings("ignore", category=DeveloperWarning)

    for k, v in args.GlobalParameters:
        arguments[k] = v
    (
        arguments["CxxCompiler"],
        arguments["CCompiler"],
        arguments["Assembler"],
        arguments["OffloadBundler"],
        arguments["HipConfig"],
    ) = validateToolchain(
        arguments["CxxCompiler"],
        arguments["CCompiler"],
        arguments["Assembler"],
        arguments["OffloadBundler"],
        ToolchainDefaults.HIP_CONFIG,
    )
    if args.NoEnumerate:
        arguments["ROCmAgentEnumeratorPath"] = False
    else:
        arguments["ROCmAgentEnumeratorPath"] = validateToolchain(
            ToolchainDefaults.DEVICE_ENUMERATOR
        )

    return arguments



################################################################################
#
# Copyright (C) 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import logging
from Tensile.SolutionStructs import Convolution
log =logging.getLogger("testlog")

"""
These tests run the convolution-vs-contraction mode always
"""

def test_simple(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)


def test_stride1x2(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Stride': '1x2',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_stride2x1(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Stride': '2x1',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_stride2x3(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Stride': '2x3',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_filter1x2(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '1x2',
                      })

    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)


def test_filter2x1(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '2x1',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_filter2x3(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '2x3',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_dilation1x2(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '2x2',
                      'Dilation': '1x2',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_dilation2x1(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '2x2',
                      'Dilation': '2x1',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_dilation2x3(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '2x2',
                      'Dilation': '2x3',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_pad1x0(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '2x2',
                      'PadStart': '1x0',
                      'PadEnd': '1x0',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_pad0x1(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '2x2',
                      'PadStart': '0x1',
                      'PadEnd': '0x1',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_pad2x3(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '2x2',
                      'PadStart': '2x3',
                      'PadEnd': '2x3',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)


def test_filter_stride_dilation_0(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'TensorBFormat': 'KCYX',
                      'TensorDFormat': 'NCHW',
                      'UnrollOnChannel': 0,
                      'Filter': '2x3',
                      'Stride': '2x3',
                      'Dilation': '2x3',
                      })
    assert(z['NumIndicesC']==4)
    assert(z['IndexAssignmentsA']==[6,5, 0,1, 4,3])
    assert(z['IndexAssignmentsB']==[6,5, 4, 2, 3])
    assert(z['UseInitialStridesAB'])
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)

def test_filter_stride_dilation_1(run_convolution_vs_contraction):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '6x7',
                      'Stride': '2x3',
                      'Dilation': '4x5',
                      'Spatial': '27x51',
                      })
    log.debug(conv.printUsage(z))
    run_convolution_vs_contraction(conv)



################################################################################
#
# Copyright (C) 2019-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from .DataType import DataType
from . import Hardware
from . import Properties
from .SolutionStructs import Solution as OriginalSolution
from .Utils import state, state_key_ordering

from . import Common

@state_key_ordering
class FreeIndex:
    StateKeys = ['isA', 'i', 'c', 'd']

    def __init__(self, isA, i=None, c=None, d=None):
        self.isA = isA
        self.i = i # index into A or B (depending on isA)
        self.c = c
        self.d = d

@state_key_ordering
class BatchIndex:
    StateKeys = ['a', 'b', 'c', 'd']
    def __init__(self, a=None, b=None, c=None, d=None):
        self.a = a
        self.b = b
        self.c = c
        self.d = d

@state_key_ordering
class BoundIndex:
    StateKeys = ['a', 'b', 'aMirror', 'bMirror']
    def __init__(self, a=None, b=None, aMirror=False, bMirror=False):
        self.a = a
        self.b = b
        self.aMirror = aMirror
        self.bMirror = bMirror


class ProblemType:
    StateKeys = ['operationIdentifier', 'aType', 'bType', 'cType', 'dType',
                 'useBeta', 'highPrecisionAccumulate', 'useInitialStridesAB', 'useInitialStridesCD', 'stridedBatched', 'f32XdlMathOp', 'stochasticRounding']
    @classmethod
    def FromOriginalState(cls, d):
        indices = [None]*d['TotalIndices']
        freeIndices  = []
        batchIndices = []
        boundIndices = []

        for i in d['IndicesSummation']:
            bi = BoundIndex(aMirror=('MirrorDimsA' in d and i in d['MirrorDimsA']),
                            bMirror=('MirrorDimsB' in d and i in d['MirrorDimsB']))
            indices[i] = bi
            boundIndices.append(bi)

        for i in range(0,d["NumIndicesC"]):
            if i in d['IndicesBatch']:
                bi = BatchIndex(c=i, d=i)
                indices[i] = bi
                batchIndices.append(bi)
            else:
                assert i in d['IndicesFree']
                if i in d['IndexAssignmentsA']:
                    fi = FreeIndex(isA=True, i=d["IndexAssignmentsA"].index(i), c=i, d=i)
                elif i in d['IndexAssignmentsB']:
                    fi = FreeIndex(isA=False, i=d["IndexAssignmentsB"].index(i), c=i, d=i)
                else:
                    raise RuntimeError("free index %u not in ia or ib"%i)
                indices[i] = fi
                freeIndices.append(fi)

        for ia, ic in enumerate(d['IndexAssignmentsA']):
            indices[ic].a = ia

        for ib, ic in enumerate(d['IndexAssignmentsB']):
            indices[ic].b = ib

        for idx in indices:
            assert idx is not None
            idxState = state(idx)
            for (key, value) in list(idxState.items()):
                assert value is not None

        rv = cls()
        rv.indices = indices
        rv.freeIndices = freeIndices
        rv.batchIndices = batchIndices
        rv.boundIndices = boundIndices
        rv.aDims = len(d['IndexAssignmentsA'])
        rv.bDims = len(d['IndexAssignmentsB'])
        rv.cDims = d['NumIndicesC']
        rv.dDims = rv.cDims

        rv.aConjugate = d['ComplexConjugateA']
        rv.bConjugate = d['ComplexConjugateB']

        srcType = DataType(d['DataType'])
        dstType = DataType(d['DestDataType']) if 'DestDataType' in d else srcType
        computeType = DataType(d['ComputeDataType']) if 'ComputeDataType' in d else dstType

        rv.aType = srcType
        rv.bType = srcType
        # for hybrid 8bit float types, we need to split the type into a_type and b_type
        if srcType.isFloat8BFloat8():
            rv.aType = DataType("F8")
            rv.bType = DataType("B8")
        elif srcType.isBFloat8Float8():
            rv.aType = DataType("B8")
            rv.bType = DataType("F8")

        # We don't expect dstType and computeType to be hybrid types for now
        rv.cType = dstType
        rv.dType = dstType
        # we already checked the src/dst/compute types are supported and well-assigned in SolutionStruct
        rv.alphaType = computeType
        rv.betaType = computeType

        rv.highPrecisionAccumulate = False
        if 'HighPrecisionAccumulate' in d:
            rv.highPrecisionAccumulate = d['HighPrecisionAccumulate']

        rv.useInitialStridesAB = False
        if 'UseInitialStridesAB' in d:
            rv.useInitialStridesAB = d['UseInitialStridesAB']
        rv.useInitialStridesCD = False
        if 'UseInitialStridesCD' in d:
            rv.useInitialStridesCD = d['UseInitialStridesCD']

        rv.stridedBatched = True
        if 'StridedBatched' in d:
          rv.stridedBatched = d['StridedBatched']

        rv.setConstStrideA = []
        if 'SetConstStrideA' in d:
            rv.setConstStrideA = d['SetConstStrideA']
        rv.setConstStrideB = []
        if 'SetConstStrideB' in d:
            rv.setConstStrideB = d['SetConstStrideB']
        rv.zeroPadA=[]
        if 'ZeroPadA' in d:
            rv.zeroPadA = d['ZeroPadA']
        rv.zeroPadB=[]
        if 'ZeroPadB' in d:
            rv.zeroPadB = d['ZeroPadB']

        rv.mirrorDimsA = d['MirrorDimsA'] if 'MirrorDimsA' in d else []
        rv.mirrorDimsB = d['MirrorDimsB'] if 'MirrorDimsB' in d else []

        rv.useBeta = True
        if 'UseBeta' in d:
            rv.useBeta = d['UseBeta']

        rv.batched = d['Batched']

        rv.f32XdlMathOp = DataType(d['F32XdlMathOp']) if 'F32XdlMathOp' in d else DataType(0)

        rv.stochasticRounding = False
        if 'StochasticRounding' in d:
            rv.stochasticRounding = d['StochasticRounding']

        return rv

    def __init__(self, freeIndices=None, batchIndices=None, boundIndices=None, aDims=None, bDims=None, cDims=None, dDims=None):
        self.convolution = None
        self.freeIndices  = freeIndices
        self.batchIndices = batchIndices
        self.boundIndices = boundIndices
        self.aDims = aDims
        self.bDims = bDims
        self.cDims = cDims
        self.dDims = dDims

    @property
    def indexNames(self):
        aNames = ['_'] * self.aDims
        bNames = ['_'] * self.bDims
        cNames = ['_'] * self.cDims

        allIndexNames = 'ijklmnopqrstuvwxyz'
        index = 0

        dNames = list([allIndexNames[index+i] for i in range(self.cDims)])
        index += len(dNames)

        sumNames = list([allIndexNames[index+i] for i in range(len(self.boundIndices))])
        index += len(sumNames)

        for free in self.freeIndices:
            if free.isA:
                aNames[free.i ] = dNames[free.d]
            else:
                bNames[free.i ] = dNames[free.d]
            cNames[free.c] = dNames[free.d]

        for batch in self.batchIndices:
            name = dNames[batch.d]
            aNames[batch.a] = name
            bNames[batch.b] = name
            cNames[batch.c] = name

        for i, bound in enumerate(self.boundIndices):
            name = sumNames[i]
            aNames[bound.a] = name.upper() if bound.aMirror else name
            bNames[bound.b] = name.upper() if bound.bMirror else name

        aNames = ''.join(aNames)
        bNames = ''.join(bNames)
        cNames = ''.join(cNames)
        dNames = ''.join(dNames)
        sumNames = ''.join(sumNames)

        return (aNames, bNames, cNames, dNames, sumNames)

    @property
    def operationIdentifier(self):
        (aNames, bNames, cNames, dNames, sumNames) = self.indexNames

        aOp = 'C' if self.aConjugate else ''
        bOp = 'C' if self.bConjugate else ''

        return '_'.join(['Contraction', sumNames,
                         'A' + aNames + aOp,
                         'B' + bNames + bOp,
                         'C' + cNames,
                         'D' + dNames])

    def placeholderStr(self, includeBatch=False, includeOperation=False, includeType=False):
        ret = ""
        if includeOperation:
            ret = self.operationIdentifier
            if not self.useBeta:
                ret += "_Beta0"
            ret += "_StridedBatched{}".format(int(self.stridedBatched))
        if includeType:
            ret += "_Type_{}{}".format(DataType(self.aType).toChar(), DataType(self.cType).toChar())
            if self.highPrecisionAccumulate:
                ret += "_HPA"

        return ret

    def predicates(self, includeBatch=False, includeOperation=False, includeType=False):
        predicates = []

        #if includeBatch and not self.batched:
        #    predicates.append(ProblemPredicate("BatchSizeEqual", index=0, value=1))

        if includeOperation:
            predicates.append(ProblemPredicate("OperationIdentifierEqual", value=self.operationIdentifier))
            if not self.useBeta:
                predicates.append(ProblemPredicate("BetaZero"))
            predicates.append(ProblemPredicate("StridedBatched", value=self.stridedBatched))

        if includeType:
            predicates.append(ProblemPredicate("TypesEqual", value=(self.aType, self.bType, self.cType, self.dType)))
            predicates.append(ProblemPredicate("HighPrecisionAccumulate", value=self.highPrecisionAccumulate))
            predicates.append(ProblemPredicate("F32XdlMathOp", value=self.f32XdlMathOp))
            predicates.append(ProblemPredicate("StochasticRounding", value=self.stochasticRounding))


        return predicates

def extractDimPredicate(cls, key, value, predicateName):
    """
    Extract the predicate for AssertStrideEqual*
    Value is a dictionary
    """
    predicates = []
    for pos,val in value.items():
        if val != -1:
            predicates.append(cls(predicateName, index=pos, value=val))
    if len(predicates) == 1:
        return predicates[0]
    elif len(predicates) > 1:
        return cls.And(predicates)

class ProblemPredicate(Properties.Predicate):
    @classmethod
    def FromOriginalKeyPair(cls, pair):
        (key, value) = pair
        if key == "AssertStrideAEqual":
            return extractDimPredicate(cls, key, value, "StrideAEqual")
        if key == "AssertStrideBEqual":
            return extractDimPredicate(cls, key, value, "StrideBEqual")
        if key == "AssertStrideCEqual":
            return extractDimPredicate(cls, key, value, "StrideCEqual")
        if key == "AssertStrideDEqual":
            return extractDimPredicate(cls, key, value, "StrideDEqual")

        if key == "AssertSizeEqual":
            return extractDimPredicate(cls, key, value, "SizeEqual")
        if key == "AssertSizeGreaterThan":
            return extractDimPredicate(cls, key, value, "SizeGreaterThan")
        if key == "AssertSizeLessThan":
            return extractDimPredicate(cls, key, value, "SizeLessThan")
        if key == "AssertSizeMultiple":
            return extractDimPredicate(cls, key, value, "SizeMultiple")

        # Arithmetic intensity assertions
        if key == "AssertAIGreaterThanEqual":
            return cls("AIGreaterThanEqual", value=value) if value > 0 else None
        if key == "AssertAILessThanEqual":
            return cls("AILessThanEqual", value=value) if value > 0 else None

        # Alpha and beta value assertions
        if key == "AssertAlphaValue":
            return cls("AlphaValue", value=str(value)) if value != False else None
        if key == "AssertBetaValue":
            return cls("BetaValue", value=str(value)) if value != False else None

        if key == "AssertCEqualsD":
            return cls("CEqualsD") if value != False else None

        # TODO - remove this when logic files have been updated
        if key == 'AssertMinApproxSize':
            return None

        if key.endswith('Multiple'):
            if value == 1:
                return None

            if key == "AssertFree0ElementMultiple":
                tag = "Free0SizeMultiple"
                index = 0
            elif key == "AssertFree1ElementMultiple":
                tag = "Free1SizeMultiple"
                index = 0
            elif key == "AssertSummationElementMultiple":
                tag = "BoundSizeMultiple"
                index = -1
            else:
                raise RuntimeError("Unknown Multiple Value: {}".format(key))

            return cls(tag, index=index, value=value)

        if key.startswith('Assert'):
            raise RuntimeError("Unknown assertion key: {}".format(key))

        if key == "Fp16AltImpl":
            return cls("Fp16AltImpl") if value != False else None

        if key == "Fp16AltImplRound":
            return cls("Fp16AltImplRound") if value != False else None

    @classmethod
    def CompoundPredicates(cls, state, problemType):
        rv = []

        if not problemType.aType.isInt8x4():
            # calculate the minimum supported free dimension size
            TLUA = state['ProblemType']['TLUA']
            TLUB = state['ProblemType']['TLUB']
            minFree0 = max(state['GlobalLoadVectorWidthA'], 1) if TLUA else 1
            minFree1 = max(state['GlobalLoadVectorWidthB'], 1) if TLUB else 1
            rv += [cls('LeadingFree0SizesGreaterOrEqual', value=minFree0)]
            rv += [cls('LeadingFree1SizesGreaterOrEqual', value=minFree1)]

        if len(state["PackedC0IndicesX"]) > 1:
          rv += [cls("CDStridesEqual")]

        if "KernelLanguage" in state:
            rv += [cls("KernelLanguageCompatible", value=state["KernelLanguage"])]

        if ('GlobalSplitU' in state) and (state['GlobalSplitU'] > 1):
            if ('_GlobalAccumulation' not in state) or (state['_GlobalAccumulation'] != 'MultipleBuffer'):
                rv += [cls("DeterministicMode", value = False)]

        if ('StreamK' in state) and (state['StreamK'] > 0) and ('StreamKAtomic' in state) and (state['StreamKAtomic'] == 1):
            # StreamKAtomic = 1 uses atomic for partial tiles
            rv += [cls("DeterministicMode", value = False)]

        # debugging: mark this set to allow the problem always runnable with PK
        if 'PersistentKernel' in state and state['PersistentKernel']:
            rv += [cls("PersistentKernelCheck")]

        if ("MatrixInstruction" in state and state["MatrixInstruction"]) or \
           ("EnableMatrixInstruction" in state and state["EnableMatrixInstruction"] is True):
            rv += [cls("ArithmeticUnitCompatible", value="MFMA")]
        else:
            rv += [cls("ArithmeticUnitCompatible", value="VALU")]

        # if bufferload is performed, we output some predication info for host side,
        # to prevent from some extremely large problems from launching and causing bufferload offset limit < 2^32
        # those cases will not satisfy the assertion thus won't use the kernel.
        # See Common.py for more details, we will need four values:
        # TODO - haven't been fully tested for FP16 and BF, need to verify the false-positive
        if 'BufferLoad' in state and state['BufferLoad'] == True:
            TLUA = state['ProblemType']['TLUA']
            TLUB = state['ProblemType']['TLUB']
            MayShiftA = TLUA and state['AssertFree0ElementMultiple'] < state['GlobalLoadVectorWidthA']
            MayShiftB = TLUB and state['AssertFree1ElementMultiple'] < state['GlobalLoadVectorWidthB']
            subrv={}
            subrv['ShiftPtrElemB'] = state['GlobalLoadVectorWidthB'] if MayShiftB else 0
            subrv['ShiftPtrElemA'] = state['GlobalLoadVectorWidthA'] if MayShiftA else 0
            subrv['DUorMT1'] = state['DepthU'] if TLUB else state['MacroTile1']
            subrv['DUorMT0'] = state['DepthU'] if TLUA else state['MacroTile0']
            # value is also a dict for better readability, client side need to handel the serialization
            rv += [cls('BufferLoadOffsetLimitCheck', value=subrv)]

        # similar check is applied for bufferstore,
        # for bufferstore offset, test if the bot-right offset < 2^32,
        # it should be StrideA*MT1, so we need to output MT1 and use the StrideA of problem in host-side for predication
        if 'BufferStore' in state and state['BufferStore'] == True:
            val = state['MacroTile1']
            if (1 in state["AssertSizeLessThan"].keys()):
                # use smaller value if array size assert is enabled
                val = min(val, state["AssertSizeLessThan"][1] - 1)
            rv += [cls('BufferStoreOffsetLimitCheck', value=val)]

        if '_GlobalAccumulation' in state and state['_GlobalAccumulation'] != None and not state["StreamK"]:
            value = state['MinKForGSU'] * state['GlobalSplitU']
            rv += [cls('GlobalSplitUCheckMinK', value=value)]

        return rv

    @classmethod
    def FromOriginalState(cls, d, problemType, morePreds=[]):
        problemTypePreds = problemType.predicates(True, True, True)
        compoundPreds = cls.CompoundPredicates(d, problemType)
        extraPreds = problemTypePreds + compoundPreds + morePreds

        predicates = [p for p in map(cls.FromOriginalKeyPair, d.items()) if p is not None] + extraPreds
        return cls.And(predicates)

class TaskPredicate(Properties.Predicate):
    @classmethod
    def FromOriginalKeyPair(cls, pair):
        (key, value) = pair
        if key == "_WorkspaceSizePerElemC" and value > 0:
            return cls("WorkspaceCheck")
        return None

    @classmethod
    def FromOriginalState(cls, d, problemType, morePreds=[]):
        predicates = [p for p in map(cls.FromOriginalKeyPair, d.items()) if p is not None]
        return cls.And(predicates)

class SizeMapping:
    StateKeys = ['workGroup',
                 'macroTile',
                 'threadTile',
                 'depthU',
                 'staggerU',
                 'globalSplitU',
                 'staggerStrideShift',
                 'workGroupMapping',
                 'packSummationDims',
                 'packBatchDims',
                 'magicDivAlg',
                 'streamK',
                 'streamKAtomic',
                 'persistentKernel',
                 'persistentKernelAlongBatch',
                 'sourceKernel',
                 'globalAccumulation',
                 'workspaceSizePerElemC',
                 'preloadKernargs'
                 ]

    @classmethod
    def FromOriginalState(cls, d):
        globalAccum = 0
        if d['_GlobalAccumulation'] == 'SingleBuffer':
            globalAccum = 1
        if d['_GlobalAccumulation'] == 'MultipleBuffer':
            globalAccum = 2
        if d['_GlobalAccumulation'] == 'PartialsBuffer':
            globalAccum = 3

        assert d["PreloadKernelArguments"] in (0, 1), d["PreloadKernelArguments"]
        return cls(workGroup             = d['WorkGroup'],
                   macroTile             = cls.ReadOriginalMacroTile(d),
                   threadTile            = d['ThreadTile'],
                   workGroupMapping      = d['WorkGroupMapping'],
                   staggerU              = d['StaggerU'] if 'StaggerU' in d else 0,
                   depthU                = d['DepthU'],
                   globalSplitU          = d['GlobalSplitU'],
                   staggerStrideShift    = d['_staggerStrideShift'] if '_staggerStrideShift' in d else 0,
                   packSummationDims     = d['PackSummationDims'] if 'PackSummationDims' in d else 0,
                   packBatchDims         = d['PackBatchDims'] if 'PackBatchDims' in d else 0,
                   streamK               = d['StreamK'] if 'StreamK' in d else 0,
                   streamKAtomic         = d['StreamKAtomic'] if 'StreamKAtomic' in d else 0,
                   persistentKernel      = d['PersistentKernel'] if 'PersistentKernel' in d else 0,
                   persistentKernelAlongBatch   = d['PersistentKernelAlongBatch'] if 'PersistentKernelAlongBatch' in d else False,
                   magicDivAlg           = d.get('MagicDivAlg', 1),
                   sourceKernel          = d['KernelLanguage'] == 'Source',
                   globalAccumulation    = globalAccum,
                   workspaceSizePerElemC = d['_WorkspaceSizePerElemC'],
                   preloadKernargs       = d["PreloadKernelArguments"]
                   )

    @classmethod
    def ReadOriginalMacroTile(cls, d):
        rv = [1,1,1]
        rv[0] = d['MacroTile0']
        rv[1] = d['MacroTile1']
        return rv

    def __init__(self, **kwargs):
        for (key, value) in list(kwargs.items()):
            setattr(self, key, value)

class Solution:
    StateKeys = ['name',
                'problemType',
                'hardwarePredicate',
                'problemPredicate',
                'taskPredicate',
                'sizeMapping',
                'debugKernel',
                'libraryLogicIndex',
                'index',
                'ideals',
                'linearModel']
    HiddenKeys = ['originalSolution']

    @classmethod
    def FromSolutionStruct(cls, solution):
        return cls.FromOriginalState(solution._state)

    @classmethod
    def FromOriginalState(cls, d, deviceInfo=None):
        rv = cls()


        if 'SolutionNameMin' in d:
            rv.name = d['SolutionNameMin']

        rv.problemType = ProblemType.FromOriginalState(d['ProblemType'])

        rv.problemPredicate = ProblemPredicate.FromOriginalState(d, rv.problemType)
        rv.taskPredicate = TaskPredicate.FromOriginalState(d, rv.problemType)

        if 'DebugKernel' in d:
            rv.debugKernel = d['DebugKernel']

        if 'SolutionIndex' in d:
            rv.index = d['SolutionIndex']

        info = cls.ReadOriginalInfo(d)
        rv.libraryLogicIndex = int(info.get("SolutionIndex", -1))

        rv.sizeMapping = SizeMapping.FromOriginalState(d)
        if 'Ideals' in d:
            rv.ideals = d['Ideals']
        else:
            rv.ideals = {}

        if 'LinearModel' in d:
            rv.linearModel = d['LinearModel']
        else:
            rv.linearModel = {}

        if 'ISA' not in d:
            if d['KernelLanguage'] == 'Assembly':
                d['ISA'] = Common.gfxArch(deviceInfo[1])
            else:
                d['ISA'] = [0,0,0]

        rv.originalSolution = OriginalSolution(d)
        # hacky, can just construct Convolution yet again?
        rv.problemType.convolution = rv.originalSolution["ProblemType"].convolution

        return rv

    @classmethod
    def ReadOriginalInfo(cls, d):
        return dict([(key, str(value)) for (key, value) in list(d.items()) if key != 'ProblemType'])

    def __init__(self, **kwargs):
        self.name = None
        self.problemType = None
        self.hardwarePredicate = Hardware.HardwarePredicate('TruePred')
        self.problemPredicate = ProblemPredicate('TruePred')
        self.taskPredicate = TaskPredicate('TruePred')
        self.sizeMapping = None
        self.debugKernel = False
        self.libraryLogicIndex = {}
        self.index = None
        self.ideals = {}

        for key, value in kwargs:
            if key not in Solution.StateKeys and key not in Solution.HiddenKeys:
                raise KeyError("{0} is not a property of Solution.".format(key))

            setattr(self, key, value)



################################################################################
#
# Copyright (C) 2016-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from .Common import architectureMap, gfxArch, printExit, printWarning, versionIsCompatible
from .SolutionStructs import Solution, ProblemSizes, ProblemType
from . import __version__
from . import SolutionLibrary
from .Utilities.ConditionalImports import yamlLoader, yamlDumper

from typing import NamedTuple, Optional

import os

try:
    import yaml
except ImportError:
    printExit(
        "You must install PyYAML to use Tensile (to parse config files). See http://pyyaml.org/wiki/PyYAML for installation instructions."
    )

try:
    import msgpack
except ImportError:
    print("Message pack python library not detected. Must use YAML backend instead.")


def updateProblemDatatypes(problemType):
    problemType["DataType"] = problemType["DataType"].value
    problemType["DestDataType"] = problemType["DestDataType"].value
    problemType["ComputeDataType"] = problemType["ComputeDataType"].value
    problemType["F32XdlMathOp"] = problemType["F32XdlMathOp"].value


###################
# Writing functions
###################
def write(filename_noExt, data, format="yaml"):
    """Writes data to file with specified format; extension is appended based on format."""
    if format == "yaml":
        writeYAML(filename_noExt + ".yaml", data)
    elif format == "msgpack":
        writeMsgPack(filename_noExt + ".dat", data)
    else:
        printExit("Unrecognized format {}".format(format))


def writeYAML(filename, data, **kwargs):
    """Writes data to file in YAML format."""
    # set default kwags for yaml dump
    if "explicit_start" not in kwargs:
        kwargs["explicit_start"] = True
    if "explicit_end" not in kwargs:
        kwargs["explicit_end"] = True
    if "default_flow_style" not in kwargs:
        kwargs["default_flow_style"] = None

    with open(filename, "w") as f:
        yaml.dump(data, f, yamlDumper, **kwargs)


def writeMsgPack(filename, data):
    """Writes data to file in Message Pack format."""
    with open(filename, "wb") as f:
        msgpack.pack(data, f)


def writeSolutions(filename, problemSizes, solutions, cache=False):
    """Writes solution YAML file."""

    # convert objects to nested dictionaries
    solutionStates = []

    if cache:
        solYaml = readYAML(filename)
        solutionStates = solYaml[2:]
    else:
        for solution in solutions:
            solutionState = solution.getAttributes()
            solutionState["ProblemType"] = solutionState["ProblemType"].state
            updateProblemDatatypes(solutionState["ProblemType"])
            solutionStates.append(solutionState)
    # write dictionaries
    with open(filename, "w") as f:
        f.write("- MinimumRequiredVersion: {}\n".format(__version__))
        f.write("- ProblemSizes:\n")
        if problemSizes:
            for sizeRange in problemSizes.ranges:
                f.write("  - Range: {}\n".format(sizeRange))
            for problemExact in problemSizes.exacts:
                #FIXME-problem, this ignores strides:
                f.write("  - Exact: {}\n".format(problemExact))

        yaml.dump(solutionStates, f, yamlDumper, default_flow_style=None)


###############################
# Reading and parsing functions
###############################
def readYAML(filename):
    """Reads and returns YAML data from file."""
    with open(filename, "r") as f:
        data = yaml.load(f, yamlLoader)
    return data


def parseSolutionsFile(filename):
    """Wrapper function to read and parse a solutions file."""
    return parseSolutionsData(readYAML(filename), filename)


def parseSolutionsData(data, srcFile="?"):
    """Parses problem sizes and solutions from the data of a solutions file."""
    if len(data) < 3:
        printExit("Solution file {} is missing required fields (len = {} < 3" \
                .format(srcFile, len(data)))

    versionString = data[0]["MinimumRequiredVersion"]
    if not versionIsCompatible(versionString):
        printWarning("Version = {} in solution file {} does not match Tensile version = {}" \
                .format(srcFile, versionString, __version__) )

    if "ProblemSizes" not in data[1]:
        printExit("Solution file {} doesn't begin with ProblemSizes".format(srcFile))

    problemSizesConfig = data[1]["ProblemSizes"]

    solutions = []
    for i in range(2, len(data)):
        solutionState = data[i]
        # force redo the deriving of parameters, make sure old version logic yamls can be validated
        solutionState["AssignedProblemIndependentDerivedParameters"] = False
        solutionState["AssignedDerivedParameters"] = False
        solutionObject = Solution(solutionState)
        solutions.append(solutionObject)
    problemType = solutions[0]["ProblemType"]
    problemSizes = ProblemSizes(problemType, problemSizesConfig)
    return (problemSizes, solutions)


class LibraryLogic(NamedTuple):
    """Return tuple for parseLibraryLogicData()"""
    schedule: str
    architecture: str
    problemType: ProblemType
    solutions: list
    exactLogic: list
    library: SolutionLibrary.MasterSolutionLibrary


def parseLibraryLogicFile(filename):
    """Wrapper function to read and parse a library logic file."""
    return parseLibraryLogicData(readYAML(filename), filename)


def parseLibraryLogicData(data, srcFile="?"):
    """Parses the data of a library logic file."""
    if type(data) is list:
        data = parseLibraryLogicList(data, srcFile)

    if "CUCount" not in data:
        data["CUCount"] = None

    if "IsAPU" not in data:
        data["IsAPU"] = None

    if "Fp16AltImpl" not in data:
        data["Fp16AltImpl"] = False

    if "Fp16AltImplRound" not in data:
        data["Fp16AltImplRound"] = False

    if "ArchitectureName" not in data:
        raise ValueError("'ArchitectureName' not found in logic file data" \
            f" Review the library logic file {srcFile}.")
    elif data["ArchitectureName"] not in architectureMap:
        raise ValueError(f"{data['ArchitectureName']} is not a supported architecture." \
            f" Review the library logic file {srcFile}.")

    if not versionIsCompatible(data["MinimumRequiredVersion"]):
        printWarning("Version = {} in library logic file {} does not match Tensile version = {}" \
                .format(srcFile, data["MinimumRequiredVersion"], __version__) )

    # unpack problemType
    problemType = ProblemType(data["ProblemType"])
    # unpack solutions
    solutions = []
    for solutionState in data["Solutions"]:
        if solutionState["ProblemType"] == "derive":
            solutionState["ProblemType"] = problemType
        if solutionState["KernelLanguage"] == "Assembly":
            # Set solution ISA based on the top-level gfx architecture in the library logic file
            solutionState["ISA"] = gfxArch(data["ArchitectureName"])
        else:
            solutionState["ISA"] = (0, 0, 0)
        # force redo the deriving of parameters, make sure old version logic yamls can be validated
        solutionState["AssignedProblemIndependentDerivedParameters"] = False
        solutionState["AssignedDerivedParameters"] = False
        solutionObject = Solution(solutionState)

        if solutionObject["ProblemType"] != problemType:
            printExit("ProblemType in library logic file {} doesn't match solution: {} != {}" \
                    .format(srcFile, problemType, solutionObject["ProblemType"]))
        solutions.append(solutionObject)

    newLibrary = SolutionLibrary.MasterSolutionLibrary.FromOriginalState(data, solutions)

    return LibraryLogic(data["ScheduleName"], data["ArchitectureName"], problemType, solutions, \
            data.get("ExactLogic"), newLibrary)


def parseLibraryLogicList(data, srcFile="?"):
    """Parses the data of a matching table style library logic file."""
    if len(data) < 9:
        printExit("Library logic file {} is missing required fields (len = {} < 9)" \
                .format(srcFile, len(data)))

    rv = {}
    rv["MinimumRequiredVersion"] = data[0]["MinimumRequiredVersion"]
    rv["ScheduleName"] = data[1]
    rv["DeviceNames"] = data[3]
    rv["ProblemType"] = data[4]
    rv["Solutions"] = data[5]

    if type(data[2]) is dict:
        rv["ArchitectureName"] = data[2]["Architecture"]
        if "CUCount" in data[2]:
            rv["CUCount"] = data[2]["CUCount"]
        if "IsAPU" in data[2]:
            rv["IsAPU"] = data[2]["IsAPU"]
    else:
        rv["ArchitectureName"] = data[2]
        rv["CUCount"] = None
        rv["IsAPU"] = None

    if architectureMap[rv["ArchitectureName"]] != rv["ScheduleName"]:
        raise ValueError(f"Architecture mismatch: {rv['ArchitectureName']} does not match " \
            f" {rv['ScheduleName']}. Review the library logic file {srcFile}.")

    rv["ExactLogic"] = data[7]
    # data[8] previously contained range logic, which has been retired

    # optional fields
    if len(data) > 10 and data[10]:
        rv["PerfMetric"] = data[10]

    if len(data) > 11 and data[11] == "Fp16AltImpl":
        rv["Fp16AltImpl"] = True

    if len(data) > 13 and data[13] == "Fp16AltImplRound":
        rv["Fp16AltImplRound"] = True

    # library logic fields
    rv["LibraryType"] = "Matching"
    rv["Library"] = {}
    rv["Library"]["indexOrder"] = data[6]
    rv["Library"]["table"] = data[7]
    rv["Library"]["distance"] = "Euclidean"
    if len(data) > 12 and data[12]:
        rv["Library"]["distance"] = data[12]

    return rv


def rawLibraryLogic(data):
    """Returns a tuple of the data in a library logic file."""
    versionString = data[0]
    scheduleName = data[1]
    architectureName = data[2]
    deviceNames = data[3]
    problemTypeState = data[4]
    solutionStates = data[5]
    indexOrder = data[6]
    exactLogic = data[7]
    rangeLogic = data[8]
    otherFields = []

    dataLength = len(data)
    if dataLength > 9:
        for idx in range(9, dataLength):
            otherFields.append(data[idx])

    return (versionString, scheduleName, architectureName, deviceNames,\
            problemTypeState, solutionStates, indexOrder, exactLogic, rangeLogic, otherFields)


########################
# Library logic creation
########################
def createLibraryLogic(schedulePrefix, architectureName, deviceNames, logicTuple, dictFormat):
    if not dictFormat:
        return createLibraryLogicList(schedulePrefix, architectureName, deviceNames, logicTuple)

    rv = {}
    rv["MinimumRequiredVersion"] = __version__
    rv["ScheduleName"] = schedulePrefix
    rv["DeviceNames"] = deviceNames
    rawProblemType = logicTuple[0]
    rawSolutions = logicTuple[1]

    # architecture
    if type(architectureName) is dict:
        rv["ArchitectureName"] = architectureName["Architecture"]
        if "CUCount" in architectureName:
            rv["CUCount"] = architectureName["CUCount"]
        if "IsAPU" in architectureName:
            rv["IsAPU"] = architectureName["IsAPU"]
    else:
        rv["ArchitectureName"] = architectureName

    # problem type
    problemTypeState = rawProblemType.state
    updateProblemDatatypes(problemTypeState)
    rv["ProblemType"] = problemTypeState

    # solutions
    solutionList = []
    for solution in rawSolutions:
        solutionState = solution.getAttributes()
        solutionState["ProblemType"] = "derive"
        solutionList.append(solutionState)

    rv["Solutions"] = solutionList

    # library logic
    exactLogic = logicTuple[3]
    exactLogicList = []
    for key in exactLogic:
        exactLogicList.append([list(key), exactLogic[key]])

    rv["LibraryType"] = "Matching"
    rv["Library"] = {}
    rv["Library"]["indexOrder"] = logicTuple[2]
    rv["Library"]["table"] = exactLogicList
    rv["Library"]["distance"] = "Euclidean"
    rv["PerfMetric"] = logicTuple[7]

    return rv


def createLibraryLogicList(schedulePrefix, architectureName, deviceNames, logicTuple):
    """Creates the data for a library logic file suitable for writing to YAML."""
    problemType = logicTuple[0]
    solutions = logicTuple[1]
    indexOrder = logicTuple[2]
    exactLogic = logicTuple[3]
    rangeLogic = logicTuple[4]

    tileSelection = False
    if len(logicTuple) > 5 and logicTuple[5]:
        tileSelection = True

    data = []
    # Tensile version
    data.append({"MinimumRequiredVersion": __version__})
    # schedule name
    data.append(schedulePrefix)  # change from Tensile to vega10
    data.append(architectureName)
    # schedule device names
    data.append(deviceNames)

    # problem type
    problemTypeState = problemType.state
    updateProblemDatatypes(problemTypeState)
    data.append(problemTypeState)

    # solutions
    solutionList = []
    for solution in solutions:
        solutionState = solution.getAttributes()
        solutionState["ProblemType"] = solutionState["ProblemType"].state
        updateProblemDatatypes(solutionState["ProblemType"])
        solutionList.append(solutionState)

    if tileSelection:
        tileSolutions = logicTuple[5]
        for solution in tileSolutions:
            solutionState = solution.getAttributes()
            solutionState["ProblemType"] = solutionState["ProblemType"].state
            updateProblemDatatypes(solutionState["ProblemType"])
            solutionList.append(solutionState)

    data.append(solutionList)
    # index order
    data.append(indexOrder)

    # exactLogic
    exactLogicList = []
    for key in exactLogic:
        exactLogicList.append([list(key), exactLogic[key]])
    data.append(exactLogicList)

    # rangeLogic
    data.append(rangeLogic)

    if tileSelection:
        tileSelectionLogic = {}
        tileSelectionIndices = logicTuple[6]
        tileSelectionLogic["TileSelectionIndices"] = tileSelectionIndices
        data.append(tileSelectionLogic)
    else:
        data.append(None)

    data.append(logicTuple[7])
    return data


def initAsmCapsCache(cacheFile: str) -> Optional[dict]:
    """
    If requested and exists, reads an assembler capabilities cache from disk.

    Arguments:
      cacheFile: Cache file (YAML format).

    Returns:
      newcache:  If cachFile is empty, None. Otherwise, if file exists, a dictionary containing
                 capabilities cache; if not, an empty dict
    """
    # if cacheFile str is empty, do not use cache
    if not cacheFile:
        return None

    cacheExists = os.path.exists(cacheFile)
    if not cacheExists:
        return {}

    cache  = readYAML(cacheFile)

    # gaurd against inconsistent state
    if not cache:
        return None

    toTuple = lambda s: tuple(int(i.strip()) for i in s.split(","))
    newcache = {toTuple(k): v for k,v in cache.items()}
    return newcache

def writeAsmCapsCache(cacheFile: str, data: dict):
    """
    Writes an asm cache to disk.

    Arguments:
      cacheFile: User spedified file to write the cache.
      data:      dict containg the computed cache.
    """
    newdata = {", ".join(map(str, k)): v for k, v in data.items()}
    writeYAML(cacheFile, newdata)



################################################################################
#
# Copyright (C) 2016-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from .Common import tPrint, HR, printExit, defaultAnalysisParameters, globalParameters, \
  setWorkingPath, popWorkingPath, assignParameterWithDefault, startTime, printWarning
from .SolutionStructs import Solution
from . import Utils
from . import LibraryIO
from . import SolutionSelectionLibrary

from copy import deepcopy
from sys import stdout

import array
import csv
import os
import time

################################################################################
# Analyze Problem Type
################################################################################
def analyzeProblemType( problemType, problemSizeGroups, inputParameters ):
  tPrint(3, HR)
  tPrint(1, "# Analyzing: %s" % problemType)

  enableTileSelection = problemType["TileAwareSelection"]
  solutionsList = []
  problemSizesList = []
  dataFileNameList = []
  selectionFileNameList = []

  for problemSizeGroup in problemSizeGroups:
    problemSizes = problemSizeGroup[0]
    dataFileName = problemSizeGroup[1]
    dataFileNameList.append(dataFileName)
    solutionsFileName = problemSizeGroup[2]
    #print "  problemSizes:", problemSizes
    #print "# DataFileName:", dataFileName
    #print "  solutionsFileName:", solutionsFileName
    if enableTileSelection:
      selectionFileName = problemSizeGroup[3]
      selectionFileNameList.append(selectionFileName)

    ######################################
    # Read Solutions
    # (problemSizes, solutions) are already read and kept in problemSizeGroups, no need to call LibraryIO.readSolutions(solutionsFileName) again
    solutions = problemSizeGroup[4]
    problemSizesList.append(problemSizes)
    solutionsList.append(solutions)
    solutionMinNaming = Solution.getMinNaming(solutions)
    tPrint(1, "# Read: %s" % (solutionsFileName))
    tPrint(3, "# ProblemSizes: %s" % problemSizes)
    tPrint(3, "# Solutions:")
    solutionIdx = 0
    for solution in solutions:
      tPrint(3, "#  (%u) %s" % (solutionIdx, Solution.getNameMin(solution, \
          solutionMinNaming)))
      solutionIdx += 1
    tPrint(3, HR)

  ######################################
  # Create Logic Analyzer
  logicAnalyzer = LogicAnalyzer( problemType, problemSizesList, solutionsList, \
      dataFileNameList, inputParameters)

  selectionSolutionsIdsList = None
  selectionSolutions = None

  validSelectionSolutions = []

  ######################################
  # Remove invalid solutions
  logicAnalyzer.removeInvalidSolutions()

  ######################################
  # Remove least important solutions
  if globalParameters["SolutionSelectionAlg"] == 0:
    logicAnalyzer.removeLeastImportantSolutions()
  elif globalParameters["SolutionSelectionAlg"] == 1:
    logicAnalyzer.keepWinnerSolutions()
  else:
    printExit("Bad KeepLogic=%u"%globalParameters["KeepLogic"])

  # print raw data
  if globalParameters["PrintLevel"] >= 3:
    line = "After Removals:\n"
    numOther = 1
    for size in logicAnalyzer.numProblemSizes:
      numOther *= size
    numCols = logicAnalyzer.numProblemSizes[1]
    if numCols == 0: numCols = 1
    numOther //= numCols
    for row in range(0, numOther):
      for col in range(0, numCols):
        for sol in range(0, logicAnalyzer.numSolutions):
         line += "% 5.0f" % logicAnalyzer.data[sol + logicAnalyzer.numSolutions*(col + row*numCols)]
        line += "; "
      line += "\n"
    print(line)

  for i in range(0, len(logicAnalyzer.solutions)):
    s = logicAnalyzer.solutions[i]
    s["SolutionIndex"] = i
    s["SolutionNameMin"] = Solution.getNameMin(s, solutionMinNaming)
    tPrint(1, "(%2u) %s : %s" % (i, Solution.getNameMin(s, solutionMinNaming), Solution.getNameFull(s)))

  if enableTileSelection:
    validSelectionSolutions = SolutionSelectionLibrary.analyzeSolutionSelection(problemType, selectionFileNameList, \
        logicAnalyzer.numSolutionsPerGroup,  logicAnalyzer.solutionGroupMap, solutionsList)

    validSelectionSolutionsIncluded = []
    validSelectionSolutionsRemainder = []
    selectionSolutionsIds = set([])
    for validSelectionSolution in validSelectionSolutions:
      (validSolution, validSolutionInfo) = validSelectionSolution
      if validSolution in logicAnalyzer.solutions:
        validExactSolutionIndex = logicAnalyzer.solutions.index(validSolution)
        selectionSolutionsIds.add(validExactSolutionIndex)
        validExactSolution = logicAnalyzer.solutions[validExactSolutionIndex]
        validSelectionSolutionsIncluded.append((validExactSolution, validSolutionInfo))
      else:
        validSelectionSolutionsRemainder.append(validSelectionSolution)

    selectionSolutions = []
    for i in range(0 ,len(validSelectionSolutionsIncluded)):
      validSelectionSolution = validSelectionSolutionsIncluded[i]
      (validSolution, validSolutionInfo) = validSelectionSolution
      validSolution["Ideals"] = validSolutionInfo

    solutionsStartIndex = len(logicAnalyzer.solutions)

    for i in range(0, len(validSelectionSolutionsRemainder)):
      validSelectionSolution = validSelectionSolutionsRemainder[i]
      (validSolution, validSolutionInfo) = validSelectionSolution
      selectionSolutionIndex = solutionsStartIndex + i
      selectionSolutionsIds.add(selectionSolutionIndex)
      validSolution["SolutionNameMin"] = Solution.getNameMin(validSolution, solutionMinNaming)
      validSolution["Ideals"] = validSolutionInfo
      selectionSolutions.append(validSolution)

    selectionSolutionsIdsList = list(selectionSolutionsIds)

  ######################################
  # Correct outliers
  """
  if inputParameters["SmoothOutliers"]:
    logicAnalyzer.smoothOutliers()
  """
  numProblemSizes = logicAnalyzer.numProblemSizes
  # print all 2D
  numPermutations = 1
  permutations = []
  for i in range(0, logicAnalyzer.numIndices):
    if i != logicAnalyzer.idx0 and i != logicAnalyzer.idx1:
      numPermutations *= numProblemSizes[i]
  #print numPermutations
  for j in range(0, numPermutations):
    pIdx = j
    permutation = []
    for i in range(0, logicAnalyzer.numIndices):
      if i != logicAnalyzer.idx0 and i != logicAnalyzer.idx1:
        npsi = numProblemSizes[i]
        permutation.append(pIdx%npsi)
        pIdx /= numProblemSizes[i]
    permutations.append(permutation)
  #print permutations
  for permutation in permutations:
    logicAnalyzer.print2D(permutation)

  ######################################
  # Range Logic
  rangeLogic = logicAnalyzer.enRule(0, logicAnalyzer.globalIndexRange)
  tPrint(3, "# Final Range Logic:")
  tPrint(3, rangeLogic)
  logicComplexity = [0]*logicAnalyzer.numIndices
  logicAnalyzer.scoreLogicComplexity(rangeLogic, logicComplexity)
  tPrint(3, "# Range Logic Complexity: %s" % logicComplexity)
  score = logicAnalyzer.scoreRangeForLogic( \
      logicAnalyzer.globalIndexRange, rangeLogic)
  tPrint(1, "\n# Score: %.0f ms" % (score/1000))
  logicAnalyzer.prepareLogic(rangeLogic) # convert indices to sizes, -1

  ######################################
  # Range Logic
  exactLogic = logicAnalyzer.exactWinners
  tPrint(1, "# Exact Logic:\n")
  tPrint(1, "%s"%exactLogic)

  #selectionSolutionsIdsList = list(selectionSolutionsIds)
  return (problemType, logicAnalyzer.solutions, logicAnalyzer.indexOrder, \
       exactLogic, rangeLogic, selectionSolutions, selectionSolutionsIdsList, logicAnalyzer.perfMetric)



################################################################################
# LogicAnalyzer
################################################################################
class LogicAnalyzer:

  ##############################################################################
  ##############################################################################
  ###
  ###  Entry / Top-Level Functions
  ###
  ##############################################################################
  ##############################################################################

  ##############################################################################
  # ENTRY: Init
  ##############################################################################
  def __init__(self, problemType, problemSizesList, solutionsList, \
      dataFileNameList, inputParameters):

    # parameters
    self.parameters = inputParameters

    # problem type
    self.problemType = problemType
    self.idx0 = self.problemType["Index0"]
    self.idx1 = self.problemType["Index1"]
    self.idxU = self.problemType["IndexUnroll"]

    # merge solutions from size groups
    # solutions needs to be a set, and offset needs to be mapping
    tPrint(1, "# Merging Solutions:")
    self.numSolutionsPerGroup = []
    self.solutionGroupMap = []
    self.solutions = []
    solutionsHash = {} # for accelerating lookups

    totalSolutions = 0
    for solutionGroupIdx in range(0, len(solutionsList)):
      solutionGroup = solutionsList[solutionGroupIdx]
      totalSolutions += len(solutionGroup)
    progressBar = Utils.ProgressBar(totalSolutions, desc="LogicAnalyzer")
    for solutionGroupIdx in range(0, len(solutionsList)):
      solutionGroup = solutionsList[solutionGroupIdx]
      self.numSolutionsPerGroup.append(len(solutionGroup))
      self.solutionGroupMap.append({})
      for solutionIdx in range(0, len(solutionGroup)):
        solution = solutionGroup[solutionIdx]
        if not solution in solutionsHash:
          sIdx = len(self.solutions) # the one we are about to add
          self.solutions.append(solution)
          solutionsHash[solution] = sIdx
        else:
          sIdx = solutionsHash[solution]

        self.solutionGroupMap[solutionGroupIdx][solutionIdx] = sIdx
        progressBar.increment()
    self.numSolutions = len(self.solutions)
    self.solutionMinNaming = Solution.getMinNaming(self.solutions)
    self.solutionNames = []
    self.solutionTiles = []
    for solution in self.solutions:
      self.solutionNames.append(Solution.getNameMin(solution, \
          self.solutionMinNaming))
      self.solutionTiles.append("%ux%u"%(solution["MacroTile0"], \
          solution["MacroTile1"]))
    self.flopsPerMac = self.problemType["DataType"].flopsPerMac()

    # merge problem sizes from size groups
    #self.numIndices = len(problemSizesList[0].numProblemSizes)
    self.numIndices = self.problemType["TotalIndices"] + problemType["NumIndicesLD"]
    unifiedProblemSizes = []
    for i in range(0, self.numIndices):
      unifiedProblemSizes.append(set())
    self.exactProblemSizes = set()
    self.rangeProblemSizes = set()
    for problemSizes in problemSizesList:
      # add exacts
      for problem in problemSizes.exacts:
        self.exactProblemSizes.add(tuple(problem.sizes))

      # add ranges
      #print "ProblemSizes", problemSizes.sizes
      #FIXME-problem
      self.rangeProblemSizes.update([tuple(problem.sizes) for problem in problemSizes.problems])
      for rangeSize in problemSizes.ranges:

        if globalParameters["ExpandRanges"]:
          # Treat ranges as pile of exacts:
          for rsize in rangeSize.problemSizes:
            self.exactProblemSizes.add(tuple(rsize))
        else:
          # Create the ranges info in the logic file
          #print "RangeSize", rangeSize
          sizedIdx = 0
          mappedIdx = 0
          for i in range(0, self.numIndices):
            if rangeSize.indexIsSized[i]:
              index = rangeSize.indicesSized[sizedIdx]
              sizedIdx += 1
            else:
              index = rangeSize.indicesSized[ \
                rangeSize.indicesMapped[mappedIdx]]
              mappedIdx += 1
            currentSize = index[0]
            currentStride = index[1]
            while currentSize <= index[3]:
              unifiedProblemSizes[i].add(currentSize)
              currentSize += currentStride
              currentStride += index[2]
    for i in range(0, len(unifiedProblemSizes)):
      unifiedProblemSizes[i] = sorted(list(unifiedProblemSizes[i]))
    tPrint(3, "UnifiedProblemSizes: %s" % unifiedProblemSizes)
    tPrint(3, "ExactProblemSizes: %s" % self.exactProblemSizes)
    tPrint(3, "RangeProblemSizes: %s" % self.rangeProblemSizes)

    # problem size index <-> size
    self.problemSizeToIndex = []
    self.problemIndexToSize = []
    self.numProblemSizes = []
    for i in range(0, self.numIndices):
      self.problemSizeToIndex.append({})
      self.problemIndexToSize.append([])
      for j in range(0, len(unifiedProblemSizes[i])):
        size = unifiedProblemSizes[i][j]
        self.problemSizeToIndex[i][size] = j
        self.problemIndexToSize[i].append(size)
      self.numProblemSizes.append(len(unifiedProblemSizes[i]))
    tPrint(1, "# NumProblemSizes: %s" % self.numProblemSizes)

    # total size of data array
    self.totalProblems = 1
    for numProblems in self.numProblemSizes:
      self.totalProblems *= numProblems
    self.totalSize = self.totalProblems * self.numSolutions
    tPrint(3, "TotalProblems: %u" % self.totalProblems)
    tPrint(3, "TotalSolutions: %u" % self.numSolutions)
    tPrint(3, "TotalSize: %u" % self.totalSize)
    # data is a 2D array [problemIdx][solutionIdx] which stores perf data in gflops for
    # the specified solution
    self.data = array.array('f', [-2]*self.totalSize)

    # Each entry in exactWinners is a 2D array [solutionIdx, perf]
    self.exactWinners = {}

    """
    # map problem sizes -> index
    self.problemSizeToIndex = []
    self.problemIndexToSize = []
    sizedIdx = 0
    mappedIdx = 0
    for i in range(0, self.numIndices):
      self.problemSizeToIndex.append({})
      self.problemIndexToSize.append([])
      if self.problemSizes.indexIsSized[i]:
        index = self.problemSizes.indicesSized[sizedIdx]
        sizedIdx += 1
      else:
        index = self.problemSizes.indicesSized[ \
          self.problemSizes.indicesMapped[mappedIdx]]
        mappedIdx += 1
      currentSize = index[0]
      currentStride = index[1]
      idx = 0
      while currentSize <= index[3]:
        self.problemSizeToIndex[i][currentSize] = idx
        self.problemIndexToSize[i].append(currentSize)
        currentSize += currentStride
        currentStride += index[2]
        idx += 1
    """
    #self.rangeIndicesFree = range(0, self.problemType["NumIndicesC"])
    #self.rangeIndicesSummation = range(self.problemType["NumIndicesC"], \
    #    self.problemType["TotalIndices"])
    self.indexOrder = self.recommendedIndexOrder()
    tPrint(3, "IndexOrder: %s" % self.indexOrder)
    self.globalIndexRange = []
    for i in range(0, self.numIndices):
      self.globalIndexRange.append([0, self.numProblemSizes[i]])
    self.problemIndicesForGlobalRange \
        = self.problemIndicesForRange(self.globalIndexRange)
    self.tab = [""]*self.numIndices

    ######################################
    # Read Data From CSV
    for fileIdx in range(0, len(dataFileNameList)):
      dataFileName = dataFileNameList[fileIdx]
      self.addFromCSV(dataFileName, self.numSolutionsPerGroup[fileIdx], \
          self.solutionGroupMap[fileIdx])



    #print self.data
    # map exact problem sizes to solutions
    tPrint(1, "# ExactWinners: %s" % self.exactWinners)


  ##############################################################################
  # ENTRY: Add From CSV
  ##############################################################################
  def addFromCSV(self, dataFileName, numSolutions, solutionMap):

    # open file
    print("reading datafile", dataFileName)
    try:
      dataFile = open(dataFileName, "r")
    except IOError:
      printExit("Can't open \"%s\" to get data" % dataFileName )

    # column indices
    csvFile = csv.reader(dataFile)
    problemSizeStartIdx = 1
    # notice that for OperationType != GEMM, the numIndices = 0
    totalSizeIdx = problemSizeStartIdx + self.numIndices

    # need to take care if the loaded csv is the export-winner-version
    csvHasWinner = "_CSVWinner" in dataFileName
    if csvHasWinner:
      # the column of the two are fixed (GFlops, SizeI/J/K/L, LDD/C/A/B, TotalFlops, WinnerGFlops, WinnerTimeUs, WinnerIdx, WinnerName)
      # the order are implemented in ResultFileReporter.cpp
      columnOfWinnerGFlops = 10
      columnOfWinnerIdx = 12

    # iterate over rows
    rowIdx = 0
    for row in csvFile:
      rowIdx+=1
      if rowIdx == 1:
        # get unit (gflops or gflops/cu) of benchmark data
        perfUnit = row[0]
        if perfUnit == "GFlops":
          self.perfMetric = "DeviceEfficiency"
        elif perfUnit == "GFlopsPerCU":
          self.perfMetric = "CUEfficiency"
        else:
          printWarning("Performance unit %s in %s is unrecognized: assuming GFlops (device efficiency)" % (perfUnit, dataFileName))
          self.perfMetric = "DeviceEfficiency"

        # get the length of each row, and derive the first column of the solution instead of using wrong "solutionStartIdx = totalSizeIdx + 1"
        rowLength = len(row)
        solutionStartIdx = rowLength - numSolutions
        continue
      else:
        #if len(row) < rowLength:
        #  printWarning("CSV File %s row %u doesn't have %u elements; ignoring remainer of file." \
        #      % (dataFileName, rowIdx, rowLength) )
        #  break

        # get problem size
        problemSize = []
        for i in range(problemSizeStartIdx, totalSizeIdx):
          problemSize.append(int(row[i]))
        problemSize = tuple(problemSize)

        # Exact Problem Size
        if problemSize in self.exactProblemSizes:

          if csvHasWinner:
            # Faster. Get the winner info from csv directly, avoid an extra loop
            winnerGFlops = float(row[columnOfWinnerGFlops])
            winnerIdx = int(row[columnOfWinnerIdx])
          else:
            # Old code. TODO - Can we get rid of this in the future?
            # solution gflops
            solutionIdx = 0
            winnerIdx = -1
            winnerGFlops = -1
            for i in range(solutionStartIdx, rowLength):
              gflops = float(row[i])
              if gflops > winnerGFlops:
                winnerIdx = solutionIdx
                winnerGFlops = gflops
              solutionIdx += 1

          if winnerIdx != -1:
            if problemSize in self.exactWinners:
              if winnerGFlops > self.exactWinners[problemSize][1]:
                #print "update exact", problemSize, "CSV index=", winnerIdx, self.exactWinners[problemSize], "->", solutionMap[winnerIdx], winnerGFlops
                self.exactWinners[problemSize] = [solutionMap[winnerIdx], winnerGFlops]
            else:
              self.exactWinners[problemSize] = [solutionMap[winnerIdx], winnerGFlops]
              #print "new exact", problemSize, "CSV index=", winnerIdx, self.exactWinners[problemSize]

        # Range Problem Size
        elif problemSize in self.rangeProblemSizes:
          problemIndices = []
          for i in range(0, self.numIndices):
            problemIndices.append(self.problemSizeToIndex[i][problemSize[i]])
          serialIdx = self.indicesToSerial(0, problemIndices)
          # solution gflops
          solutionIdx = 0
          for i in range(solutionStartIdx, rowLength):
            gflops = float(row[i])
            self.data[serialIdx+solutionMap[solutionIdx]] = gflops
            solutionIdx += 1

        # Unknown Problem Size
        else:
          printExit("Huh? %s has ProblemSize %s which isn't in its yaml" \
              % ( dataFileName, list(problemSize)) )
    #print self.data


  ##############################################################################
  # ENTRY: Remove Invalid Solutions
  ##############################################################################
  def removeInvalidSolutions(self):
    #problemIndices = [0]*self.numIndices
    allSolutionValid = False
    while not allSolutionValid:
      invalidIdx = -1
      for problemIndices in self.problemIndicesForGlobalRange:
        problemSerial = self.indicesToSerial(0, problemIndices)
        for solutionIdx in range(0, self.numSolutions):
          gflops = self.data[problemSerial+solutionIdx]
          if gflops == 0:
            invalidIdx = solutionIdx
            break
      if invalidIdx >= 0:
        tPrint(1, "# Removing Invalid Solution: %u %s" \
            % (invalidIdx, self.solutionNames[invalidIdx]) )
        self.removeSolution(invalidIdx)
      else:
        allSolutionValid = True


  ##############################################################################
  # ENTRY: Original KeepLogic algorithm: Remove Least Important Solutions,
  # one at a time.  Stop when leastImportantSolution indicates no more
  # solutions can be removed, which appears to be when the solution
  # is used by an exact problem or is the only possible solution for some
  # problem or doesn't improve the a solution by > SolutionImportanceMin%
  ##############################################################################
  def removeLeastImportantSolutions(self):
    # Remove least important solutions
    start = time.time()
    while len(self.solutions) > 1:
      lisTuple = self.leastImportantSolution()
      if lisTuple != None:
        lisIdx = lisTuple[0]
        lisPercSaved = lisTuple[1]
        lisPercWins = lisTuple[2]
        lisPercTime = lisTuple[3]
        if lisPercSaved < self.parameters["SolutionImportanceMin"] or lisPercWins == 0:
          tPrint(1, "# Removing Unimportant Solution %u/%u: %s ( %f%% wins, %f%% ms time, %f%% ms saved" \
              % (lisIdx, self.numSolutions, self.solutionNames[lisIdx], 100*lisPercWins, 100*lisPercTime, 100*lisPercSaved) )
          self.removeSolution(lisIdx)
          continue
        else:
          break
      else: # no more lis, remainders are exact winner
        break
    stop = time.time()
    print("removeLeastImportantSolutions elapsed time = %.1f secs" % (stop - start))


  ##############################################################################
  # ENTRY: Alternate KeepLogic algorithm that keeps the fastest for each
  #  exact and range.  Other solutions are removed.
  ##############################################################################
  def keepWinnerSolutions(self):

    # solution indexes for the winners:
    winners = set()

    solutionImportance = []
    for i in range(0, self.numSolutions):
      solutionImportance.append([i, 0, 0, 0, False])
    problemSizes = [0]*self.numIndices
    print("problemIndicesForGlobalRange", self.problemIndicesForGlobalRange)
    for problemIndices in self.problemIndicesForGlobalRange:
      for i in range(0, self.numIndices):
        problemSizes[i] = self.problemIndexToSize[i][problemIndices[i]]
      totalFlops = self.flopsPerMac
      for size in problemSizes:
        totalFlops *= size

      problemSerial = self.indicesToSerial(0, problemIndices)
      winnerIdx = -1
      winnerGFlops = -1e6
      for solutionIdx in range(0, self.numSolutions):
        solutionSerialIdx = problemSerial + solutionIdx
        solutionGFlops = self.data[solutionSerialIdx]
        if solutionGFlops > winnerGFlops:
          winnerIdx = solutionIdx
          winnerGFlops = solutionGFlops

      winners.add(winnerIdx)

    # Always keep the exact sizes:
    for exactProblem in self.exactWinners:
      winnerIdx = self.exactWinners[exactProblem][0]
      #print "keepWinnerSolution adding exact", exactProblem, winnerIdx
      winners.add(winnerIdx)

    print("Winners", winners)
    self.pruneSolutions(winners)



  ##############################################################################
  # ENTRY: En Rule
  # currentIndexIndex = 0, 1, 2, 3...
  # currentIndexRange will have only 1 size for prior indices (unless initial)
  #
  # Rule:
  # [128, [
  #         [64, [
  #                [16, 0],
  #                [2880,1]
  #              ]
  #         ],
  #         [96, [
  #                [16, 0],
  #                [64, 1]
  #              ]
  #         ]
  #       ]
  # ], another
  #
  #
  ##############################################################################
  def enRule(self, currentIndexIndex, currentIndexRange):
    cii = currentIndexIndex
    if currentIndexIndex == 0:
      self.tab[cii] = "[] "
    elif currentIndexIndex == 1:
      self.tab[cii] = "[%2u] " % ( \
          currentIndexRange[self.indexOrder[0]][0])
    elif currentIndexIndex == 2:
      self.tab[cii] = "[%2u,%2u] " % ( \
          currentIndexRange[self.indexOrder[0]][0], \
          currentIndexRange[self.indexOrder[1]][0])
    elif currentIndexIndex == 3:
      self.tab[cii] = "[%2u,%2u,%2u] " % ( \
          currentIndexRange[self.indexOrder[0]][0], \
          currentIndexRange[self.indexOrder[1]][0], \
          currentIndexRange[self.indexOrder[2]][0])
    elif currentIndexIndex == 4:
      self.tab[cii] = "[%2u,%2u,%2u,%2u] " % ( \
          currentIndexRange[self.indexOrder[0]][0], \
          currentIndexRange[self.indexOrder[1]][0], \
          currentIndexRange[self.indexOrder[2]][0], \
          currentIndexRange[self.indexOrder[3]][0])
    tab = self.tab[cii]
    if globalParameters["PrintLevel"] == 1:
      stdout.write("\n%s"%tab)
    currentIndex = self.indexOrder[currentIndexIndex]
    tPrint(3, "%senRule(%s)" % (tab, currentIndexRange))
    nextIndexIndex = currentIndexIndex+1
    nextIndexRange = deepcopy(currentIndexRange)
    isLastIndex = currentIndexIndex == self.numIndices-1
    ruleList = []

    ########################################
    # SingleProblem
    ########################################
    if currentIndexRange[currentIndex][1] \
        - currentIndexRange[currentIndex][0] == 1:

      ########################################
      # SingleProblem & LastIndex
      # this is last index, so just return fastest solution
      ########################################
      if isLastIndex:
        tPrint(3, "%sSingleProblem & LastIndex" % tab)
        winnerIdx = self.winnerForRange(currentIndexRange)
        if winnerIdx < 0:
          tPrint(3, "%sSingleProblem & LastIndex :: winnerIdx<0; returning" % (tab) )
          return None
        ruleList.append([-1, winnerIdx])
        if globalParameters["PrintLevel"] == 1:
          stdout.write("%")

      ########################################
      # SingleProblem & NotLastIndex
      # this isn't last index, so just recursively return next index
      ########################################
      else:
        tPrint(3, "%sSingleProblem & NotLastIndex" % tab)
        #    % (tab, nextIndexRange) )
        nextRule = self.enRule(nextIndexIndex, nextIndexRange)
        if nextRule == None:
          tPrint(3, "%sSingleProblem & NotLastIndex :: nextRule==None; returning" % (tab) )
          return None
        rule = [ -1, nextRule ]
        ruleList.append(rule)
        if globalParameters["PrintLevel"] == 1:
          stdout.write("%")

    else:
    ########################################
    # MultiProblem
    # Create Initial Rule
    ########################################

      if isLastIndex:
        ########################################
        # MultiProblem & LastIndex
        # InitialRule using winnerForRange()
        ########################################
        tPrint(3, "%sMultiProblem & LastIndex" % tab)
        winnerIdx = -1
        for problemIndex in range(currentIndexRange[currentIndex][0], \
            currentIndexRange[currentIndex][1]):
          nextIndexRange[currentIndex][0] = problemIndex
          nextIndexRange[currentIndex][1] = problemIndex+1
          winnerIdx = self.winnerForRange(nextIndexRange)
          initialRule = [ currentIndexRange[currentIndex][0], winnerIdx]
          if winnerIdx >= 0:
            break
        if winnerIdx < 0:
          tPrint(3, "%sMultiProblem & LastIndex :: winnerIdx<0; returning" % (tab) )
          return None

      else:
        ########################################
        # MultiProblem & NotLastIndex
        # InitialRule using enRule()
        ########################################
        tPrint(3, "%sMultiProblem & NotLastIndex" % tab)

        # create initial rule
        winnerIdx = -1
        nextRule = None
        for problemIndex in range(currentIndexRange[currentIndex][0], \
            currentIndexRange[currentIndex][1]):
          nextIndexRange[currentIndex][0] = problemIndex
          nextIndexRange[currentIndex][1] = problemIndex+1
          nextRule = self.enRule(nextIndexIndex, nextIndexRange)
          # break when found initial rule
          if nextRule != None:
            break

        if nextRule == None:
          printWarning("%sMultiProblem & NotLastIndex :: nextRule==None; returning" % (tab) )
          return None
        initialRule = [ currentIndexRange[currentIndex][0], nextRule ]
      ruleList.append(initialRule)
      tPrint(3, "%sMultiProblem::InitialRuleList=%s" % (tab, ruleList))
      if globalParameters["PrintLevel"] == 1:
        stdout.write("#")

      ########################################
      # MultiProblem
      # Append Rules to Initial Rule
      ########################################
      tPrint(3, "%sMultiProblem::Improving Rule" % tab)
      for problemIndex in range(currentIndexRange[currentIndex][0]+1, \
          currentIndexRange[currentIndex][1]):
        nextIndexRange[currentIndex][0] = problemIndex
        nextIndexRange[currentIndex][1] = problemIndex+1
        priorRule = ruleList[len(ruleList)-1]
        priorRuleForSize = deepcopy(priorRule)
        priorRuleForSize[0] = problemIndex

        if isLastIndex:
          ########################################
          # nextRule using winnersForRange()
          winnerIdx = self.winnerForRange(nextIndexRange)
          tPrint(3, "%sMultiProblem::ImproveRule[%u]::LastIndex::WinnerIdx=%u for %s" % (tab, problemIndex, winnerIdx, nextIndexRange))
          # if no solutions benchmarked for this problem size, continue
          if winnerIdx < 0:
            ruleList[len(ruleList)-1][0] = problemIndex # NO_UPDATE
            tPrint(3, "%sUpdating range b/c None" % tab)
            if globalParameters["PrintLevel"] == 1:
              stdout.write(" ")
            continue
          else:
            candidateRule = [ problemIndex, winnerIdx]
        else:
          ########################################
          # nextRule using enRule()
          nextRule = self.enRule(nextIndexIndex, nextIndexRange)
          tPrint(3, "%sMultiProblem::ImproveRule[%u]::NotLastIndex::NextRule=%s for %s; %s" % (tab, problemIndex, nextRule, nextIndexIndex, nextIndexRange))
          if nextRule == None:
            ruleList[len(ruleList)-1][0] = problemIndex # NO_UPDATE
            tPrint(3, "%sUpdating b/c None" % tab)
            if globalParameters["PrintLevel"] == 1:
              stdout.write(" ")
            continue
          else:
            candidateRule = [ problemIndex, nextRule ]

        ########################################
        # candidate same as prior
        if candidateRule[1] == priorRule[1]:
          tPrint(3, "%sCandidateRule==PriorRule; just updating prior" % (tab))
          ruleList[len(ruleList)-1][0] = problemIndex # NO_UPDATE
          if globalParameters["PrintLevel"] == 1:
            stdout.write(" ")
          continue

        ########################################
        # compare candidate vs prior
        else:
          tPrint(3, "%sCandidateRule!=PriorRule; appending rule assuming its better" % (tab))

          """
          priorRuleScore = self.scoreRangeForLogic(nextIndexRange, \
              [priorRuleForSize])
          logicComplexity = [0]*self.numIndices
          self.scoreLogicComplexity( \
              [priorRuleForSize], logicComplexity)
          priorRuleScore += self.parameters["BranchPenalty"] \
              * sum(logicComplexity)
          # score candidate
          candidateRuleScore = self.scoreRangeForLogic(nextIndexRange, \
              [candidateRule])
          #print "CRS", candidateRuleScore
          logicComplexity = [0]*self.numIndices
          self.scoreLogicComplexity( \
              [candidateRule], logicComplexity)
          candidateRuleScore += self.parameters["BranchPenalty"] \
              * sum(logicComplexity)
          candidateRuleScore += self.parameters["BranchPenalty"] # penalize
          candidateFaster = candidateRuleScore < priorRuleScore
          tPrint(3, "%sP[%2u]: %s %s~%.0fus < %s~%.0fus" % (tab, problemIndex, \
              "wins" if candidateFaster else "same", \
              candidateRule, candidateRuleScore, priorRuleForSize, \
              priorRuleScore ))
          """

          ########################################
          # candidate wins
          #print candidateRuleScore, priorRuleScore
          if True: # or candidateRuleScore < priorRuleScore:
            ruleList.append(candidateRule)
            tPrint(3, "%sAppending b/c Different" % tab)
            if globalParameters["PrintLevel"] == 1:
              stdout.write("#")

          ########################################
          # prior wins
          else:
            tPrint(3, "%sPrior Rule Wins" % tab)
            if globalParameters["PrintLevel"] == 1:
              stdout.write(".")
            ruleList[len(ruleList)-1][0] = problemIndex # NO_UPDATE

    tPrint(3, "%sReturning RuleList: %s" % (tab, ruleList))
    return ruleList



  ##############################################################################
  ##############################################################################
  ###
  ###  Mid-Level Functions
  ###
  ##############################################################################
  ##############################################################################



  ##############################################################################
  # Prepare Logic
  # convert threshold indices to sizes
  # last threshold = -1
  ##############################################################################
  def prepareLogic(self, logic):
    depth = self.getLogicDepth(logic)
    if depth == 0: return
    indexIndex = self.numIndices - depth
    index = self.indexOrder[indexIndex]
    for i in range(0, len(logic)):
      if i == len(logic)-1:
        logic[i][0] = -1
      else:
        logic[i][0] = self.problemIndexToSize[index][logic[i][0]]
      self.prepareLogic(logic[i][1])


  ##############################################################################
  # Print2D
  ##############################################################################
  def print2D(self, indices ):
    indicesIdx = 0
    problemIndices = []
    for i in range(0, self.numIndices):
      if i == self.idx0:
        problemIndices.append(-1)
      elif i == self.idx1:
        problemIndices.append(-1)
      else:
        problemIndices.append(indices[indicesIdx])
        indicesIdx += 1

    winnerIndices = []
    w = "winner"
    g = "gflops"
    f = "faster"
    s = "second"
    sss = []
    for sIdx in range(0, self.numSolutions):
      sss.append("Sol[%u]" % sIdx)
    for j in range(0, self.numProblemSizes[1]):
      w += ",%4u" % self.problemIndexToSize[1][j]
      g += ",%4u" % self.problemIndexToSize[1][j]
      f += ",%4u" % self.problemIndexToSize[1][j]
      s += ",%4u" % self.problemIndexToSize[1][j]
      for sIdx in range(0, self.numSolutions):
        sss[sIdx] += ",%4u" % self.problemIndexToSize[1][j]
    w += "\n"
    g += "\n"
    f += "\n"
    s += "\n"
    for sIdx in range(0, self.numSolutions):
      sss[sIdx] += "\n"
    for i in range(0, self.numProblemSizes[0]):
      problemIndices[self.idx0] = i
      w += "%4u" % self.problemIndexToSize[0][i]
      g += "%4u" % self.problemIndexToSize[0][i]
      f += "%4u" % self.problemIndexToSize[0][i]
      s += "%4u" % self.problemIndexToSize[0][i]
      for sIdx in range(0, self.numSolutions):
        sss[sIdx] += "%4u" % self.problemIndexToSize[0][i]
      for j in range(0, self.numProblemSizes[1]):
        problemIndices[self.idx1] = j
        problemSerial = self.indicesToSerial(0, problemIndices)
        for sIdx in range(0, self.numSolutions):
          sss[sIdx] += ",%f" % self.data[problemSerial+sIdx]
        winnerIdx = 0
        secondIdx = 1
        winnerGFlops = self.data[problemSerial+0]
        secondGFlops = 1e-9
        for solutionIdx in range(1, self.numSolutions):
          solutionSerialIdx = problemSerial + solutionIdx
          solutionGFlops = self.data[solutionSerialIdx]
          if solutionGFlops > winnerGFlops:
            secondIdx = winnerIdx
            secondGFlops = winnerGFlops
            winnerIdx = solutionIdx
            winnerGFlops = solutionGFlops
          elif solutionGFlops > secondGFlops:
            secondIdx = solutionIdx
            secondGFlops = solutionGFlops

        if winnerIdx not in winnerIndices:
          winnerIndices.append(winnerIdx)
        w += ",%4u" % winnerIdx
        g += ",%f" % winnerGFlops
        f += ",%f" % (winnerGFlops/secondGFlops)
        s += ",%4u" % (secondIdx)
      w += "\n"
      g += "\n"
      f += "\n"
      s += "\n"
      for sIdx in range(0, self.numSolutions):
        sss[sIdx] += "\n"

    w += "\n\n"
    g += "\n\n"
    f += "\n\n"
    s += "\n\n"
    for sIdx in range(0, self.numSolutions):
      sss[sIdx] += "\n\n"
    w += "Winners:\n"
    for winnerIdx in winnerIndices:
      w += "%4u, %s, %s\n" % (winnerIdx, self.solutionTiles[winnerIdx], self.solutionNames[winnerIdx])

    printFileName = "Winner2D"
    for idx in indices:
      printFileName += "_%u" % idx
    printFileName += ".csv"
    printFile = open(os.path.join(globalParameters["WorkingPath"], printFileName), "w")
    printFile.write( w )
    printFile.write( g )
    printFile.write( f )
    printFile.write( s )
    for sIdx in range(0, self.numSolutions):
      printFile.write( sss[sIdx] )
    printFile.close()


  ##############################################################################
  # Least Important Solution
  ##############################################################################
  def leastImportantSolution(self):
    solutionImportance = []
    for i in range(0, self.numSolutions):
      solutionImportance.append([i, 0, 0, 0, False])
    problemSizes = [0]*self.numIndices
    totalSavedMs = 0
    totalExecMs = 0
    totalWins = 0
    for problemIndices in self.problemIndicesForGlobalRange:
      for i in range(0, self.numIndices):
        problemSizes[i] = self.problemIndexToSize[i][problemIndices[i]]
      totalFlops = self.flopsPerMac
      for size in problemSizes:
        totalFlops *= size

      problemSerial = self.indicesToSerial(0, problemIndices)
      winnerIdx = -1
      winnerGFlops = -1e6
      secondGFlops = -1e9
      for solutionIdx in range(0, self.numSolutions):
        solutionSerialIdx = problemSerial + solutionIdx
        solutionGFlops = self.data[solutionSerialIdx]
        if solutionGFlops > winnerGFlops:
          secondGFlops = winnerGFlops
          winnerIdx = solutionIdx
          winnerGFlops = solutionGFlops
        elif solutionGFlops > secondGFlops:
          secondGFlops = solutionGFlops

      winnerTimeMs = totalFlops / winnerGFlops / 1000000.0
      secondTimeMs = totalFlops / secondGFlops / 1000000.0
      if winnerGFlops > 0 and secondGFlops > 0:
        solutionImportance[winnerIdx][1] += (secondTimeMs - winnerTimeMs)
        totalSavedMs += secondTimeMs - winnerTimeMs
      if winnerGFlops > 0:
        solutionImportance[winnerIdx][2] += 1
        solutionImportance[winnerIdx][3] += winnerTimeMs
        totalExecMs += winnerTimeMs
        totalWins += 1
        if secondGFlops <= 0:
          solutionImportance[winnerIdx][4] = True # this is only valid solution for this problem size, keep it


    # print data before sorting
    for i in range(0, self.numSolutions):
      tPrint(3, "[%2u] %s: %e saved, %u wins, %u time, %s" \
          % (solutionImportance[i][0], \
          self.solutionNames[solutionImportance[i][0]], \
          solutionImportance[i][1], \
          solutionImportance[i][2], \
          solutionImportance[i][3], \
          "singular" if solutionImportance[i][4] else "" ) )

    totalSavedMs = max(1, totalSavedMs)
    solutionImportance.sort(key=lambda x: x[1])
    for i in range(0, self.numSolutions):
      solutionIdx = solutionImportance[i][0]
      canRemove = not solutionImportance[i][4] # don't remove if is only win for any size
      for exactProblem in self.exactWinners:
        winnerIdx = self.exactWinners[exactProblem][0]
        if solutionIdx == winnerIdx: # exact winners are important
          canRemove = False
          break
      if canRemove:
        idx = solutionImportance[i][0]
        if totalSavedMs > 0:
          percSaved = 1.0 * solutionImportance[i][1] / totalSavedMs
        else:
          percSaved = 0
        if totalWins > 0:
          percWins = 1.0 * solutionImportance[i][2] / totalWins
        else:
          percWins = 0
        if totalExecMs > 0:
          percTime = 1.0 * solutionImportance[i][3] / totalExecMs
        else:
          percTime = 0
        return ( idx, percSaved, percWins, percTime )
    return None


  ##############################################################################
  # Remove Solution
  ##############################################################################
  def removeSolution(self, removeSolutionIdx):

    # temporarily move current to old
    oldSolutions = deepcopy(self.solutions)
    oldNumSolutions = self.numSolutions
    oldData = deepcopy(self.data)

    # update solutions
    self.solutions = []
    for i in range(0, oldNumSolutions):
      if i != removeSolutionIdx:
        self.solutions.append(oldSolutions[i])
    self.solutionMinNaming = Solution.getMinNaming(self.solutions)
    self.solutionNames = []
    self.solutionTiles = []
    for solution in self.solutions:
      self.solutionNames.append(Solution.getNameMin(solution, \
          self.solutionMinNaming))
      self.solutionTiles.append("%ux%u"%(solution["MacroTile0"], \
          solution["MacroTile1"]))
    self.numSolutions = len(self.solutions)

    # update data
    self.totalSize = self.totalProblems * self.numSolutions
    self.data = array.array('f', [0]*self.totalSize)
    for problemIndex in range(0, self.totalProblems):
      newSolutionIdx = 0
      for oldSolutionIdx in range(0, oldNumSolutions):
        if oldSolutionIdx != removeSolutionIdx:
          self.data[problemIndex*self.numSolutions+newSolutionIdx] \
              = oldData[problemIndex*oldNumSolutions+oldSolutionIdx]
          newSolutionIdx += 1

    # update exact Winners
    for problemSize in self.exactWinners:
      if self.exactWinners[problemSize][0] >= removeSolutionIdx:
        self.exactWinners[problemSize][0] -= 1


  ##############################################################################
  # Prune a list of solutions, keeping only the indices specified in
  # keepSolutions.  keepSolutions is a set not a list
  ##############################################################################
  def pruneSolutions(self, keepSolutions):

    removeSolutionIdxList = []
    solutionMapNewToOld = [] # dense mapping
    solutionMapOldToNew = [-1] * self.numSolutions

    # temporarily move current to old
    oldSolutions = deepcopy(self.solutions)
    oldNumSolutions = self.numSolutions
    oldData = deepcopy(self.data)
    # update solutions
    self.solutions = []
    for i in range(0, oldNumSolutions):
      if i in keepSolutions:
        solutionMapNewToOld.append(i)
        solutionMapOldToNew[i] = len(self.solutions)
        self.solutions.append(oldSolutions[i])
      else:
        removeSolutionIdxList.append(i)

    self.solutionMinNaming = Solution.getMinNaming(self.solutions)
    self.solutionNames = []
    self.solutionTiles = []
    for solution in self.solutions:
      self.solutionNames.append(Solution.getNameMin(solution, \
          self.solutionMinNaming))
      self.solutionTiles.append("%ux%u"%(solution["MacroTile0"], \
          solution["MacroTile1"]))
    self.numSolutions = len(self.solutions)

    # update data
    self.totalSize = self.totalProblems * self.numSolutions
    self.data = array.array('f', [0]*self.totalSize)
    for problemIndex in range(0, self.totalProblems):
      for newSolutionIdx in range(0, self.numSolutions):
        oldSolutionIdx = solutionMapNewToOld[newSolutionIdx]
        self.data[problemIndex*self.numSolutions+newSolutionIdx] \
            = oldData[problemIndex*oldNumSolutions+oldSolutionIdx]

    # update exact Winners
    for problemSize in self.exactWinners:
      #print "prune updating exacWinner", problemSize, \
      #        "from ", self.exactWinners[problemSize][0], \
      #        "to ", solutionMapOldToNew[self.exactWinners[problemSize][0]]
      self.exactWinners[problemSize][0] = \
          solutionMapOldToNew[self.exactWinners[problemSize][0]]
      if self.exactWinners[problemSize][0] == -1:
        print(("warning: exactWinner[", problemSize, "] == -1"))
      if self.exactWinners[problemSize][0] >= self.numSolutions:
        print(("warning: exactWinner[", problemSize, "] "))


  ##############################################################################
  # Score Range For Logic
  ##############################################################################
  def scoreRangeForLogic(self, indexRange, logic):
    depth = self.getLogicDepth(logic)
    depth = self.numIndices - depth
    fullLogic = deepcopy(logic)
    for i in range(0, depth):
      fullLogic = [[-1, fullLogic]]
    return self.scoreRangeForFullLogic(depth, indexRange, fullLogic)

  ##############################################################################
  # Score Range For Full Logic
  ##############################################################################
  def scoreRangeForFullLogic(self, depth, indexRange, logic):
    score = 0
    for problemIndices in self.problemIndicesForRange(indexRange):
      problemSerial = self.indicesToSerial(0, problemIndices)
      totalFlops = self.totalFlopsForProblemIndices(problemIndices)
      solutionIdx = self.getSolutionForProblemIndicesUsingLogic( \
          problemIndices, logic)
      if solutionIdx == None:
        printWarning("SolutionIdx = None. This should never happen.")
        continue
      solutionGFlops = self.data[problemSerial + solutionIdx]
      solutionGFlops = max(1E-9, solutionGFlops)
      timeUs = totalFlops / solutionGFlops / 1000
      score += timeUs
    return score

  ##############################################################################
  # Get Solution For Problem Indices Using Logic
  ##############################################################################
  def getSolutionForProblemIndicesUsingLogic(self, problemIndices, logic):
    currentProblemIndices = self.toIndexOrder(problemIndices)
    currentLogic = logic
    for i in range(0, self.numIndices):
      currentSizeIndex = currentProblemIndices[0]
      for j in range(0, len(currentLogic)):
        if currentLogic[j][0] < 0:
          currentProblemIndices = currentProblemIndices[1:]
          currentLogic = currentLogic[j][1]
          break
        if currentLogic[j][0] >= 0:
          if currentSizeIndex <= currentLogic[j][0]:
            currentProblemIndices = currentProblemIndices[1:]
            currentLogic = currentLogic[j][1]
            break
    return currentLogic


  ##############################################################################
  ##############################################################################
  ###
  ###  Helper / Low-Level Functions
  ###
  ##############################################################################
  ##############################################################################


  ##############################################################################
  # Get Winner For Problem
  def getWinnerForProblem(self, problemIndices):
    problemSerial = self.indicesToSerial(0, problemIndices)
    winnerIdx = -1
    winnerGFlops = -1
    for solutionIdx in range(0, self.numSolutions):
      solutionSerialIdx = problemSerial + solutionIdx
      solutionGFlops = self.data[solutionSerialIdx]
      solutionGFlops = max(1E-9, solutionGFlops)
      if solutionGFlops > winnerGFlops:
        winnerIdx = solutionIdx
        winnerGFlops = solutionGFlops
    # print "Winner %u %f" % (winnerIdx, winnerGFlops)
    return (winnerIdx, winnerGFlops)


  ##############################################################################
  # Winner For Range, -1 if nothing benchmarked
  def winnerForRange(self, indexRange):
    if self.numSolutions == 1:
      return 0
    else:
      scores = self.scoreRangeForSolutions(indexRange)
      winnerIdx = 0
      hasWinner = False
      for solutionIdx in range(1, self.numSolutions):
        if scores[solutionIdx] < scores[winnerIdx]:
          winnerIdx = solutionIdx
          hasWinner = True
        elif scores[solutionIdx] > scores[winnerIdx]:
          hasWinner = True
        else:
          pass # still no winner

      return winnerIdx if hasWinner else -1


  ##############################################################################
  # Score (microseconds) Range For Solutions
  def scoreRangeForSolutions(self, indexRange):
    scores = [0]*self.numSolutions
    for problemIndices in self.problemIndicesForRange(indexRange):
      problemSerial = self.indicesToSerial(0, problemIndices)
      totalFlops = self.totalFlopsForProblemIndices(problemIndices)
      for solutionIdx in range(0, self.numSolutions):
        gflops = self.data[problemSerial+solutionIdx]
        if gflops > 0:
          timeUs = totalFlops / gflops / 1000
        else: # this solution not benchmarked for this size, therefore set
          # its score to +inf so that its automatically disqualified
          timeUs = float("inf")
        scores[solutionIdx] += timeUs
    return scores


  ##############################################################################
  # Score Logic Complexity
  def scoreLogicComplexity(self, logic, logicComplexity):
    depth = self.getLogicDepth(logic)
    if depth == 0: return
    depth = self.numIndices - depth
    for i in range(0, len(logic)):
      logicComplexity[depth] += 1
      self.scoreLogicComplexity(logic[i][1], logicComplexity)


  ##############################################################################
  # Get Logic Depth
  def getLogicDepth(self, logic):
    obj = logic
    depth = 0
    while isinstance(obj, list):
      obj = obj[0][1]
      depth += 1
    return depth


  ##############################################################################
  # To Index Order
  def toIndexOrder(self, problemIndices):
    ordered = []
    for i in self.indexOrder:
      ordered.append(problemIndices[i])
    return ordered


  ##############################################################################
  # Total Flops For Problem Indices
  def totalFlopsForProblemIndices(self, problemIndices):
    totalFlops = self.flopsPerMac
    for i in range(0, self.numIndices):
      totalFlops *= self.problemIndexToSize[i][problemIndices[i]]
    return totalFlops


  ##############################################################################
  # Recommended Index Order
  # TODO, this may depend on transposes
  def recommendedIndexOrder(self):
    order = []
    for i in range(0, self.problemType["TotalIndices"]):
      if i != self.idxU and i != self.idx1 and i != self.idx0:
        order.append(i)
    order.append(self.idxU)
    order.append(self.idx0)
    order.append(self.idx1)
    return order

  ##############################################################################
  # Problem Indices For Range
  def problemIndicesForRange(self, indexRange):
    problemIndexList = []
    # early return for empty set
    for i in range(0, self.numIndices):
      if indexRange[i][0] == indexRange[i][1]:
        return []
    problemIndices = []
    for idx in indexRange:
      problemIndices.append(idx[0])
    moreProblems = True
    while moreProblems:
      problemIndexList.append(deepcopy(problemIndices))
      # next problem
      problemIndices[0] += 1
      for i in range(0, self.numIndices):
        if problemIndices[i] >= indexRange[i][1]:
          if i == self.numIndices-1:
            moreProblems = False
            break
          else:
            problemIndices[i] = indexRange[i][0]
            problemIndices[i+1] += 1
        else:
          break
    return problemIndexList


  ##############################################################################
  # Get Item
  def __getitem__(self, indexTuple):
    indices = indexTuple[0] # in analysis order
    solutionIdx = indexTuple[1]
    serial = self.indicesToSerial(solutionIdx, indices)
    return self.data[serial]


  ##############################################################################
  # Set Item
  def __setitem__(self, indexTuple, value):
    indices = indexTuple[0] # in analysis order
    solutionIdx = indexTuple[1]
    serial = self.indicesToSerial(solutionIdx, indices )
    self.data[serial] = value


  ##############################################################################
  # Indices -> Serial
  def indicesToSerial(self, solutionIdx, indices ):
    serial = 0
    stride = 1
    serial += solutionIdx * stride
    stride *= self.numSolutions
    for i in range(0, self.numIndices):
      serial += indices[i] * stride
      stride *= self.numProblemSizes[i]
    return serial



def generateLogic(config, benchmarkDataPath, libraryLogicPath):
  tPrint(3, "# LibraryLogic config: %s" % config)
  tPrint(3, "# DefaultAnalysisParameters: " % defaultAnalysisParameters)

  setWorkingPath(libraryLogicPath)

  # Assign Defaults
  analysisParameters = {}
  for parameter in defaultAnalysisParameters:
    assignParameterWithDefault(analysisParameters, parameter, config, \
        defaultAnalysisParameters)

  tPrint(1, "")
  tPrint(1, HR)
  currentTime = time.time()
  elapsedTime = currentTime - startTime
  tPrint(1, "# Analysing data in %s - %.3fs" % (globalParameters["BenchmarkDataPath"], elapsedTime) )
  for parameter in analysisParameters:
    tPrint(3, "#   %s: %s" % (parameter, analysisParameters[parameter]))
  tPrint(1, HR)
  tPrint(1, "")

  ##############################################################################
  # Determine Which Problem Types
  ##############################################################################
  problemTypes = {}
  if not os.path.exists(benchmarkDataPath):
    printExit("Path doesn't exist: %s" % benchmarkDataPath)
  fileNames = os.listdir(benchmarkDataPath)
  fileNames = sorted(fileNames)
  for fileName in fileNames:
    if os.path.splitext(fileName)[1] == ".csv":
      fileBase = os.path.splitext( \
          os.path.join(benchmarkDataPath, \
          fileName))[0]
      dataFileName = fileBase + ".csv"
      solutionsFileName = fileBase + ".yaml"
      selectionFileName = fileBase + ".gsp"
      if not os.path.exists(dataFileName):
        printExit("%s doesn't exist for %s" % (dataFileName, fileBase) )
      if not os.path.exists(solutionsFileName):
        printExit("%s doesn't exist for %s" % (solutionsFileName, fileBase) )
      (problemSizes, solutions) = LibraryIO.parseSolutionsFile(solutionsFileName)
      if len(solutions) == 0:
        printExit("%s doesn't contains any solutions." % (solutionsFileName) )
      problemType = solutions[0]["ProblemType"]
      if problemType not in problemTypes:
        problemTypes[problemType] = []
      problemTypes[problemType].append( (problemSizes, \
          dataFileName, solutionsFileName, selectionFileName, solutions) )

  for problemType in problemTypes:
    logicTuple = analyzeProblemType(problemType, problemTypes[problemType], analysisParameters)

    filename = os.path.join(globalParameters["WorkingPath"], \
        "{}_{}".format(analysisParameters["ScheduleName"], str(problemType) + ".yaml"))

    tPrint(3, "# writing library logic YAML {}".format(filename))
    dictFormat = globalParameters["DictLibraryLogic"]
    data = LibraryIO.createLibraryLogic(analysisParameters["ScheduleName"], \
        analysisParameters["ArchitectureName"], analysisParameters["DeviceNames"], logicTuple, dictFormat)

    LibraryIO.writeYAML(filename, data, explicit_start=False, explicit_end=False)

  currentTime = time.time()
  elapsedTime = currentTime - startTime
  tPrint(1, "%s\n# Finish Analysing data to %s in %.3fs\n%s" % (HR, os.path.split(libraryLogicPath)[0], elapsedTime, HR) )
  popWorkingPath()


################################################################################
################################################################################
###
###   Main
###
################################################################################
################################################################################
def main(  config ):

  benchmarkDataPath = os.path.join(globalParameters["WorkingPath"], \
      globalParameters["BenchmarkDataPath"])

  libraryLogicPath = os.path.join(globalParameters["WorkingPath"], \
      globalParameters["LibraryLogicPath"])

  generateLogic(config, benchmarkDataPath, libraryLogicPath)



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import logging,pytest
from Tensile.SolutionStructs import Convolution,ConvProblem
log =logging.getLogger("testlog")

def test_stride2x3():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
                          config={'TensorAFormat': 'NCHW',
                                  'Stride': '2x3',
                           })
    log.debug(conv.printUsage(z))
    e= { 'n':64, 'c':256, 'h':20, 'w':14, 'k':1024, 'x':1, 'y':1, 'u':2, 'v':3 }
    ec = ConvProblem(e, conv)
    assert (ec.sizes == (5, 10, e['k'], e['n'], e['c']))
    assert (ec.stridesA == (3, 28, e['h'] * e['w'], 71680))

def test_stride2x3_defaults():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
                          config={'TensorAFormat': 'NCHW',
                                  'Stride': '2x3',
                           })
    log.debug(conv.printUsage(z))
    e= { 'n':64, 'c':256, 'h':20, 'w':14, 'k':1024}
    ec = ConvProblem(e, conv)
    assert (ec.sizes == (5, 10, e['k'], e['n'], e['c']))
    assert (ec.stridesA == (3, 28, e['h'] * e['w'], 71680))

@pytest.mark.skip(reason="no X filter, ZeroPadA has one entry and it is for Y filter")
def test_pad_4x1():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
                          config={'TensorAFormat': 'NCHW',
                                  'Filter': '4x1',
                                  'PadStart': 'Nx0',
                                  'PadEnd'  : 'Nx0',
                           })
    log.debug(conv.printUsage(z))
    e= {'n': 1, 'c': 4, 'h': 12, 'w': 8, 'k': 1, 'p': 2 , 'p_': 3}
    ec = ConvProblem(e, conv)
    assert (ec.zeroPadA == [[1,4, 2, 3]])

def test_pad_4x3():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
                          config={'TensorAFormat': 'NCHW',
                                  'Filter': '4x3',
                                  'PadStart': 'NxN',
                                  'PadEnd'  : 'NxN',
                           })
    log.debug(conv.printUsage(z))
    e= {'n': 1, 'c': 4, 'h': 12, 'w': 8, 'k': 1, 'p': 1 , 'p_': 2, 'q': 3, 'q_': 4 }
    ec = ConvProblem(e, conv)
    assert (ec.zeroPadA[0] == [0,5, 3, 4])
    assert (ec.zeroPadA[1] == [1,4, 8, 16])
    assert (len(ec.zeroPadA) == 2)

def test_bad_filter():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
                          config={'TensorAFormat': 'NCHW',
                           })
    log.debug(conv.printUsage(z))
    e= { 'n':64, 'c':256, 'h':20, 'w':14, 'k':1024, 'x':2, 'y':1, 'u':1, 'v':1 }
    with pytest.raises(RuntimeError, \
            #match="Mismatch between ConvolutionConfig value \(1\) and specified value \(2\) for filter[0]."):
            match="Mismatch between ConvolutionConfig value"):
        ConvProblem(e, conv)


def test_bad_stride():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
                          config={'TensorAFormat': 'NCHW',
                                  'Stride': '2x3',
                           })
    log.debug(conv.printUsage(z))
    e= { 'n':64, 'c':256, 'h':20, 'w':14, 'k':1024, 'x':1, 'y':1, 'u':7, 'v':2 }
    with pytest.raises(RuntimeError, \
            match="Mismatch between ConvolutionConfig value.*stride"):
        ConvProblem(e, conv)

def test_bad_missing_field():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
                          config={'TensorAFormat': 'NCHW',
                                  'Stride': '2x3',
                           })
    log.debug(conv.printUsage(z))
    e= { 'c':256, 'h':20, 'w':14, 'k':1024, 'x':1, 'y':1, 'u':1, 'v':1 }
    with pytest.raises(Exception, match="required ConvProblem field 'n' not present in ConvProblem"):
        ConvProblem(e, conv)

def test_bad_invalid_field():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
                          config={'TensorAFormat': 'NCHW',
                                  'Stride': '2x3',
                           })
    log.debug(conv.printUsage(z))
    e= { 'A': 123, 'n':64, 'c':256, 'h':20, 'w':14, 'k':1024, 'x':1, 'y':1, 'u':1, 'v':1 }
    with pytest.raises(Exception, match="unknown ConvProblem field 'A'"):
        ConvProblem(e, conv)

def test_bad_invalid_list():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
                          config={'TensorAFormat': 'NCHW',
                                  'Stride': '2x3',
                           })
    log.debug(conv.printUsage(z))
    e= [1,2,3,4]
    with pytest.raises(Exception, match="ConvProblem must be a dictionary"):
        ConvProblem(e, conv)



################################################################################
#
# Copyright (C) 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from __future__ import print_function
import pytest
from Tensile.DataType import DataType

def test_init_single():
    expected = DataType('S')
    assert DataType('single') == expected
    assert DataType('Float') == expected
    assert DataType('tensileDataTypeFloat') == expected

def test_init_double():
    expected = DataType('D')
    assert DataType('double') == expected
    assert DataType('Double') == expected
    assert DataType('tensileDataTypeDouble') == expected

def test_init_complexSingle():
    expected = DataType('C')
    assert DataType('complexSingle') == expected
    assert DataType('complexFloat') == expected
    assert DataType('tensileDataTypeComplexFloat') == expected

def test_init_complexDouble():
    expected = DataType('Z')
    assert DataType('complexDouble') == expected
    assert DataType('complexDouble') == expected
    assert DataType('tensileDataTypeComplexDouble') == expected

def test_init_half():
    expected = DataType('H')
    assert DataType('half') == expected
    assert DataType('Half') == expected
    assert DataType('tensileDataTypeHalf') == expected

def test_init_i8():
    expected = DataType('4xi8')
    assert DataType('int8x4') == expected
    assert DataType('Int8x4') == expected
    assert DataType('tensileDataTypeInt8x4') == expected

def test_init_i32():
    expected = DataType('I')
    assert DataType('int32') == expected
    assert DataType('Int32') == expected
    assert DataType('tensileDataTypeInt32') == expected

def test_single():
    obj = DataType(0)
    assert obj.toChar() == 'S'
    assert obj.toName() == 'single'
    assert obj.toEnum() == 'Float'
    assert obj.toHIP() == 'float'
    assert obj.toDevice("") == 'float'
    assert obj.toCpp() == 'float'
    assert obj.getLibString() == 'tensileDataTypeFloat'
    assert obj.numBytes() == 4
    assert obj.isReal()

def test_double():
    obj = DataType(1)
    assert obj.toChar() == 'D'
    assert obj.toName() == 'double'
    assert obj.toEnum() == 'Double'
    assert obj.toHIP() == 'double'
    assert obj.toDevice("") == 'double'
    assert obj.toCpp() == 'double'
    assert obj.getLibString() == 'tensileDataTypeDouble'
    assert obj.numBytes() == 8
    assert obj.isReal()

def test_complexSingle():
    obj = DataType(2)
    assert obj.toChar() == 'C'
    assert obj.toName() == 'complexSingle'
    assert obj.toEnum() == 'ComplexFloat'
    assert obj.toHIP() == 'TensileComplexFloat'
    assert obj.toDevice("") == 'TensileComplexFloat'
    assert obj.toCpp() == 'TensileComplexFloat'
    assert obj.getLibString() == 'tensileDataTypeComplexFloat'
    assert obj.numBytes() == 8
    assert not obj.isReal()

def test_complexDouble():
    obj = DataType(3)
    assert obj.toChar() == 'Z'
    assert obj.toName() == 'complexDouble'
    assert obj.toEnum() == 'ComplexDouble'
    assert obj.toHIP() == 'TensileComplexDouble'
    assert obj.toDevice("") == 'TensileComplexDouble'
    assert obj.toCpp() == 'TensileComplexDouble'
    assert obj.getLibString() == 'tensileDataTypeComplexDouble'
    assert obj.numBytes() == 16
    assert not obj.isReal()

def test_half():
    obj = DataType(4)
    assert obj.toChar() == 'H'
    assert obj.toName() == 'half'
    assert obj.toEnum() == 'Half'
    assert obj.toHIP() == 'tensile_half'
    assert obj.toDevice("") == 'tensile_half'
    assert obj.toCpp() == 'TensileHalf'
    assert obj.getLibString() == 'tensileDataTypeHalf'
    assert obj.numBytes() == 2
    assert obj.isReal()

def test_int8():
    obj = DataType(5)
    assert obj.toChar() == '4xi8'
    assert obj.toName() == 'int8x4'
    assert obj.toEnum() == 'Int8x4'
    assert obj.toHIP() == 'uint32_t'
    assert obj.toDevice("") == 'uint32_t'
    assert obj.toCpp() == 'TensileInt8x4'
    assert obj.getLibString() == 'tensileDataTypeInt8x4'
    assert obj.numBytes() == 4
    assert obj.isReal()

def test_int32():
    obj = DataType(6)
    assert obj.toChar() == 'I'
    assert obj.toName() == 'int32'
    assert obj.toEnum() == 'Int32'
    assert obj.toHIP() == 'int32_t'
    assert obj.toDevice("") == 'int32_t'
    assert obj.toCpp() == 'TensileInt32'
    assert obj.getLibString() == 'tensileDataTypeInt32'
    assert obj.numBytes() == 4
    assert obj.isReal()

def test_cmp():
    assert DataType('single') == DataType('S')
    assert not DataType('S') != DataType(0)
    assert DataType('Float') < DataType('Double')
    assert not DataType('tensileDataTypeFloat') > DataType('Z')
    assert DataType('half') >= DataType('ComplexFloat')
    assert not DataType('int32') <= DataType('tensileDataTypeInt8x4')

def test_bounds():
    with pytest.raises(Exception):
        DataType(14)



################################################################################
#
# Copyright (C) 2019-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from math import log

########################################
# Format Instruction
########################################

def inst(*args):
    # exclude the last parameter (before comment)
    # if it is empty (needed for clang++ assembler)
    if len(args) > 2 and args[len(args)-2] == "":
        params = args[0:len(args)-2]
    else:
        params = args[0:len(args)-1]
    comment = args[len(args)-1]
    formatting = "%s"
    if len(params) > 1:
        formatting += " %s"
    for _ in range(0, len(params)-2):
        formatting += ", %s"
    instStr = formatting % (params)
    line = "%-50s // %s\n" % (instStr, comment)
    return line

########################################
# Format Trailing Comment Only
########################################

def instCommentOnly(comment=""):
    # Aligned with inst (50 chars)
    return "%-50s // %s\n" % ("", comment)

########################################
# Format GPRs
########################################

def gpr(*args):
    gprType = args[0]
    args = args[1]
    if isinstance(args[0], int):
        if len(args) == 1:
            return "%s%u"%(gprType, args[0])
        elif len(args) == 2:
            if args[1] == 1:
                return "%s%u"%(gprType, args[0])
            else:
                return "%s[%u:%u]"%(gprType, args[0], args[0]+args[1]-1)
    if isinstance(args[0], str):
        if len(args) == 1:
            return "%s[%sgpr%s]"%(gprType, gprType, args[0])
        elif len(args) == 2:
            if args[1] == 1:
                return "%s[%sgpr%s]"%(gprType, gprType, args[0])
            else:
                return "%s[%sgpr%s:%sgpr%s+%u]"%(gprType, gprType, args[0], \
                        gprType, args[0], args[1]-1)

def vgpr(*args):
    return gpr("v", args)

def sgpr(*args):
    return gpr("s", args)

def accvgpr(*args):
    return gpr("acc", args)

########################################
# Log 2
########################################

def log2(x):
    return int(log(x, 2) + 0.5)

########################################
# Divide & Remainder
# quotient register, remainder register, dividend register, divisor, tmpSgpr
########################################

def vectorStaticDivideAndRemainder(qReg, rReg, dReg, divisor, tmpSgpr, doRemainder=True, comment=""):

    dComment = "%s = %s / %s"    % (vgpr(qReg), vgpr(dReg), divisor) if (comment=="") else comment
    rComment = "%s = %s %% %s" % (vgpr(rReg), vgpr(dReg), divisor) if (comment=="") else comment

    kStr = ""
    if ((divisor & (divisor - 1)) == 0): # pow of 2
        # does not work with doRemainder and (qReg==dReg)
        assert (not (doRemainder and (qReg == dReg)))
        divisor_log2 = log2(divisor)
        kStr += inst("v_lshrrev_b32", vgpr(qReg), divisor_log2, vgpr(dReg), dComment)
        if doRemainder:
            kStr += inst("v_and_b32", vgpr(rReg), (divisor-1), vgpr(dReg), rComment)
    else:
        """
        if divisor == 30:
            shift = 32+2
        elif divisor >= 14:
            shift = 32+4
        elif divisor >= 7:
            shift = 32+3
        elif divisor >= 6:
            shift = 32+2 # this was 32+3 but divisor hex didn't fit into 32 bits
        elif divisor >= 5:
            shift = 32+2
        elif divisor >= 3:
            shift = 32+1
        """
        # does not work with doRemainder and (qReg==dReg or rReg==dReg or qReg==rReg)
        assert (not (doRemainder and (qReg == dReg or rReg == dReg or qReg==rReg)))
        shift = 32+1
        shiftMinus32 = shift - 32
        magic = ((2**shift) // divisor) + 1
        kStr += inst("s_mov_b32", sgpr(tmpSgpr), hex(magic), dComment)
        kStr += inst("v_mul_hi_u32", vgpr(qReg), vgpr(dReg), sgpr(tmpSgpr), dComment)
        kStr += inst("v_lshrrev_b32", vgpr(qReg), hex(shiftMinus32), vgpr(qReg), dComment)
        if doRemainder:
            kStr += inst("s_mov_b32", sgpr(tmpSgpr), hex(divisor), rComment)
            kStr += inst("v_mul_lo_u32", vgpr(rReg), vgpr(qReg), sgpr(tmpSgpr), rComment)
            kStr += inst("_v_sub_u32", vgpr(rReg), vgpr(dReg), vgpr(rReg), rComment)
    return kStr

def vectorStaticDivide(qReg, dReg, divisor, tmpSgpr, comment=""):
    rReg = -1 # unused
    kStr = vectorStaticDivideAndRemainder(qReg, rReg, dReg, divisor, tmpSgpr, False, comment)
    return kStr

def vectorStaticRemainder(rReg, dReg, divisor, tmpSgpr, comment=""):
    if comment == "":
        comment = "%s = %s %% %s" % (vgpr(rReg), vgpr(dReg), divisor)

    kStr = ""
    if ((divisor & (divisor - 1)) == 0): # pow of 2
        kStr += inst("v_and_b32", vgpr(rReg), (divisor-1), vgpr(dReg), comment)
    else:
        """
        if divisor == 30:
            shift = 32+2
        elif divisor >= 14:
            shift = 32+4
        elif divisor >= 7:
            shift = 32+3
        elif divisor >= 6:
            shift = 32+2 # this was 32+3 but divisor hex didn't fit into 32 bits
        elif divisor >= 5:
            shift = 32+2
        elif divisor >= 3:
            shift = 32+1
        """
        # does not work with qReg==rReg
        assert (rReg != dReg)
        shift = 32+1
        shiftMinus32 = shift - 32
        magic = ((2**shift) // divisor) + 1
        kStr += inst("s_mov_b32", sgpr(tmpSgpr), hex(magic), comment)
        kStr += inst("v_mul_hi_u32", vgpr(rReg), vgpr(dReg), sgpr(tmpSgpr), comment)
        kStr += inst("v_lshrrev_b32", vgpr(rReg), hex(shiftMinus32), vgpr(rReg), comment)
        kStr += inst("s_mov_b32", sgpr(tmpSgpr), hex(divisor), comment)
        kStr += inst("v_mul_lo_u32", vgpr(rReg), vgpr(rReg), sgpr(tmpSgpr), comment)
        kStr += inst("_v_sub_u32", vgpr(rReg), vgpr(dReg), vgpr(rReg), comment)
    return kStr

# only used for loop unroll and GlobalSplitU and XCC mapping
# doRemainder==0 : compute quotient only
# doRemainder==1 : compute quotient and remainder
# doRemainder==2 : only compute remainder (not quotient unless required for remainder)
# dreg == dividend
# tmpSgpr must be 2 SPGRs (can be None if divisor is power of 2)
# qReg and dReg can be "sgpr[..]" or names of sgpr (will call sgpr)
def scalarStaticDivideAndRemainder(qReg, rReg, dReg, divisor, tmpSgpr, \
        doRemainder=1):

    qRegSgpr = qReg if type(qReg) == str and qReg.startswith("s[") else sgpr(qReg)

    dRegSgpr = dReg if type(dReg) == str and dReg.startswith("s[") else sgpr(dReg)

    kStr = ""
    if ((divisor & (divisor - 1)) == 0): # pow of 2
        divisor_log2 = log2(divisor)
        if doRemainder != 2:
            kStr += inst("s_lshr_b32", qRegSgpr, dRegSgpr, divisor_log2, \
                    "%s = %s / %u"%(qRegSgpr, dRegSgpr, divisor) )
        if doRemainder:
            kStr += inst("s_and_b32", sgpr(rReg), (divisor-1), dRegSgpr, \
                    "%s = %s %% %u"%(sgpr(rReg), dRegSgpr, divisor) )
    else:
        # Temp register required if divisor is not power of 2
        assert qReg != tmpSgpr
        assert tmpSgpr != None
        """
        if divisor == 30:
            shift = 32+2
        elif divisor >= 14:
            shift = 32+4
        elif divisor >= 6:
            shift = 32+3
        elif divisor >= 5:
            shift = 32+2
        elif divisor >= 3:
            shift = 32+1
        """
        shift = 32+1
        magic = ((2**shift) // divisor) + 1
        magicHi = magic // (2**16)
        magicLo = magic & (2**16-1)

        kStr += inst("s_mov_b32", sgpr(tmpSgpr+1), hex(0), "STATIC_DIV: divisior=%s"%divisor)
        kStr += inst("s_mul_i32", sgpr(tmpSgpr+0), hex(magicHi), dRegSgpr, "tmp1 = dividend * magic hi")
        kStr += inst("s_lshl_b64", sgpr(tmpSgpr,2), sgpr(tmpSgpr,2), hex(16), "left shift 16 bits")
        kStr += inst("s_mul_i32", qRegSgpr, dRegSgpr, hex(magicLo), "tmp0 = dividend * magic lo")
        kStr += inst("s_add_u32", sgpr(tmpSgpr+0), qRegSgpr, sgpr(tmpSgpr+0), "add lo")
        kStr += inst("s_addc_u32", sgpr(tmpSgpr+1), sgpr(tmpSgpr+1), hex(0), "add hi")
        kStr += inst("s_lshr_b64", sgpr(tmpSgpr,2), sgpr(tmpSgpr,2), hex(shift), "tmp1 = (dividend * magic) << shift")
        kStr += inst("s_mov_b32", qRegSgpr, sgpr(tmpSgpr), "quotient")
        if doRemainder:
            kStr += inst("s_mul_i32", sgpr(tmpSgpr), qRegSgpr, hex(divisor), "quotient*divisor")
            kStr += inst("s_sub_u32", sgpr(rReg), dRegSgpr, sgpr(tmpSgpr), "rReg = dividend - quotient*divisor")
    return kStr

########################################
# Multiply
# product register, operand register, multiplier
########################################

# vgpr operand only version
# support multiplier < 1
def vectorStaticMultiply(product, operand, multiplier, tmpSgpr=None, comment=""):
    if comment == "":
        comment = "%s = %s * %s" % (vgpr(product), vgpr(operand), multiplier)

    if multiplier == 0:
            return inst("v_mov_b32", vgpr(product), hex(multiplier), comment)
    elif multiplier < 1:
            # use division if multiplier < 1
            multiplier = int(1/multiplier)
            return vectorStaticDivide(product, operand, multiplier, tmpSgpr, comment)
    elif ((multiplier & (multiplier - 1)) == 0): # pow of 2
        multiplier_log2 = log2(multiplier)
        if multiplier_log2==0 and product == operand:
            return instCommentOnly(comment + " (multiplier is 1, do nothing)")
        else:
            return inst("v_lshlrev_b32", vgpr(product), hex(multiplier_log2), vgpr(operand), comment)
    else:
        kStr = ""
        if product == operand:
            kStr += inst("s_mov_b32", tmpSgpr, hex(multiplier), comment)
            kStr += inst("v_mul_lo_u32", vgpr(product), tmpSgpr, vgpr(operand), comment)
        else:
            kStr += inst("v_mov_b32", vgpr(product), hex(multiplier), comment)
            kStr += inst("v_mul_lo_u32", vgpr(product), vgpr(product), vgpr(operand), comment)
        return kStr


def staticMultiply(product, operand, multiplier, tmpSgpr=None, comment=""):
    if comment == "":
        comment = "%s = %s * %s" % (product, operand, multiplier)

    if multiplier == 0:
            return inst("v_mov_b32", product, hex(multiplier), comment)
    elif ((multiplier & (multiplier - 1)) == 0): # pow of 2
        multiplier_log2 = log2(multiplier)
        if multiplier_log2==0 and product == operand:
            return instCommentOnly(comment + " (multiplier is 1, do nothing)")
        else:
            return inst("v_lshlrev_b32", product, hex(multiplier_log2), operand, comment)
    else:
        kStr = ""
        if product == operand:
            kStr += inst("s_mov_b32", tmpSgpr, hex(multiplier), comment)
            kStr += inst("v_mul_lo_u32", product, tmpSgpr, operand, comment)
        else:
            kStr += inst("v_mov_b32", product, hex(multiplier), comment)
            kStr += inst("v_mul_lo_u32", product, product, operand, comment)
        return kStr


########################################
# Multiply scalar for 64bit
# product register, operand register, multiplier
########################################

def scalarStaticMultiply(product, operand, multiplier, tmpSgpr=None, comment=""):
    if comment == "":
        comment = "%s = %s * %s" % (product, operand, multiplier)

    if multiplier == 0:
            return inst("s_mov_b64", product, hex(multiplier), comment)

    # TODO- to support non-pow2, need to use mul_32 and mul_hi_32 ?
    assert ((multiplier & (multiplier - 1)) == 0) # assert pow of 2

    multiplier_log2 = log2(multiplier)
    if multiplier_log2==0 and product == operand:
        return instCommentOnly(comment + " (multiplier is 1, do nothing)")
    else:
        # notice that the src-order of s_lshl_b64 is different from v_lshlrev_b32.
        return inst("s_lshl_b64", product, operand, hex(multiplier_log2), comment)
